"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""


from sregistry.logger import bot
from subprocess import Popen, PIPE, STDOUT
import os

################################################################################
# Software Versions
################################################################################


def get_singularity_version(singularity_version=None):
    """get_singularity_version will determine the singularity version for a
    build first, an environmental variable is looked at, followed by
    using the system version.

    Parameters
    ==========
    singularity_version: if not defined, look for in environment. If still
    not find, try finding via executing --version to Singularity. Only return
    None if not set in environment or installed.
    """

    if singularity_version is None:
        singularity_version = os.environ.get("SINGULARITY_VERSION")

    if singularity_version is None:
        try:
            cmd = ["singularity", "--version"]
            output = run_command(cmd)

            if isinstance(output["message"], bytes):
                output["message"] = output["message"].decode("utf-8")
            singularity_version = output["message"].strip("\n")
            bot.info("Singularity %s being used." % singularity_version)

        except:
            singularity_version = None
            bot.warning("Singularity version not found, so it's likely not installed.")

    return singularity_version


def which(software=None, strip_newline=True):
    """get_install will return the path to where Singularity (or another
    executable) is installed.
    """
    if software is None:
        software = "singularity"
    cmd = ["which", software]
    try:
        result = run_command(cmd)
        if strip_newline is True:
            result["message"] = result["message"].strip("\n")
        return result

    except:  # FileNotFoundError
        return None


def check_install(software=None, quiet=True):
    """check_install will attempt to run the singularity command, and
    return True if installed. The command line utils will not run
    without this check.

    Parameters
    ==========
    software: the software to check if installed
    quiet: should we be quiet? (default True)
    """
    if software is None:
        software = "singularity"
    cmd = [software, "--version"]
    try:
        version = run_command(cmd, software)
    except:  # FileNotFoundError
        return False
    if version is not None:
        if quiet is False and version["return_code"] == 0:
            version = version["message"]
            bot.info("Found %s version %s" % (software.upper(), version))
        return True
    return False


def get_installdir():
    """get_installdir returns the installation directory of the application"""
    return os.path.abspath(os.path.dirname(os.path.dirname(__file__)))


def get_thumbnail():
    """return the robot.png thumbnail from the database folder.
    if the user has exported a different image, use that instead.
    """
    from sregistry.defaults import SREGISTRY_THUMBNAIL

    if SREGISTRY_THUMBNAIL is not None:
        if os.path.exists(SREGISTRY_THUMBNAIL):
            return SREGISTRY_THUMBNAIL
    return "%s/database/robot.png" % get_installdir()


def run_command(cmd, sudo=False):
    """run_command uses subprocess to send a command to the terminal.

    Parameters
    ==========
    cmd: the command to send, should be a list for subprocess
    error_message: the error message to give to user if fails,
    if none specified, will alert that command failed.

    """
    if sudo is True:
        cmd = ["sudo"] + cmd

    try:
        output = Popen(cmd, stderr=STDOUT, stdout=PIPE)

    except FileNotFoundError:
        cmd.pop(0)
        output = Popen(cmd, stderr=STDOUT, stdout=PIPE)

    t = output.communicate()[0], output.returncode
    output = {"message": t[0], "return_code": t[1]}

    if isinstance(output["message"], bytes):
        output["message"] = output["message"].decode("utf-8")

    return output


def confirm_action(question, force=False):
    """confirm if the user wants to perform a certain action

    Parameters
    ==========
    question: the question that will be asked
    force: if the user wants to skip the prompt
    """
    if force is True:
        return True

    response = input(question)
    while len(response) < 1 or response[0].lower().strip() not in "ynyesno":
        response = input("Please answer yes or no: ")

    if response[0].lower().strip() in "no":
        return False

    return True


def confirm_delete(filename, force=False):
    """confirm if the user wants to delete a file

    Parameters
    ==========
    filename: the file that will be removed
    force: if the user wants to skip the prompt
    """
    return confirm_action("Are you sure you want to remove {}?".format(filename), force)



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
from sregistry.utils import parse_image_name, remove_uri
import os


def add(
    self,
    image_path=None,
    image_uri=None,
    image_name=None,
    url=None,
    metadata=None,
    save=True,
    copy=False,
):
    """dummy add simple returns an object that mimics a database entry, so the
    calling function (in push or pull) can interact with it equally. Most
    variables (other than image_path) are not used."""

    # We can only save if the image is provided
    if image_path is not None:
        if not os.path.exists(image_path):
            bot.exit("Cannot find %s" % image_path)

    if image_uri is None:
        bot.exit("You must provide an image uri <collection>/<namespace>")

    names = parse_image_name(remove_uri(image_uri))
    bot.debug("Added %s to filesystem" % names["uri"])

    # Create a dummy container on the fly
    class DummyContainer:
        def __init__(self, image_path, client_name, url, names):
            self.image = image_path
            self.client = client_name
            self.url = url
            self.name = names["image"]
            self.tag = names["tag"]
            self.uri = names["uri"]

    container = DummyContainer(image_path, self.client_name, url, names)

    bot.info("[container][add] %s" % names["uri"])
    return container


def init_db(self, db_path=None):
    """initialize the database, meaning we just set the name to be dummy"""

    # Database Setup, use default if uri not provided
    self.database = "dummy"



"""

sregistry.api: base template for making a connection to an API

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.main.base.auth import require_secrets, auth_flow

from sregistry.main.base.headers import get_headers, reset_headers, update_headers

from sregistry.main.base.http import (
    call,
    delete,
    download,
    get,
    head,
    healthy,
    paginate_get,
    post,
    put,
    stream,
    stream_response,
    verify,
)

from sregistry.main.base.inspect import get_metadata

from sregistry.main.base.settings import (
    get_setting,
    get_settings,
    get_storage_name,
    get_and_update_setting,
    required_get_and_update,
    update_setting,
)

from sregistry.logger import bot
from sregistry.defaults import SREGISTRY_DATABASE
import os


class ApiConnection(object):
    # Setup

    def __init__(self):
        self.headers = None
        self.base = None
        self._reset_headers()

        # If client initialized with _init_db, do it
        if hasattr(self, "_init_db"):
            self._init_db(SREGISTRY_DATABASE)

    # Metadata

    def _client_tagged(self, tags):
        """ensure that the client name is included in a list of tags. This is
        important for matching builders to the correct client. We exit
        on fail.

        Parameters
        ==========
        tags: a list of tags to look for client name in

        """

        # We must match the client to a tag
        name = self.client_name.lower()
        tags = [t.lower() for t in tags]

        if name not in tags:
            bot.exit("%s not found in %s, must match!" % (name, tags))

    def speak(self):
        """
        a function for the client to announce him or herself, depending
        on the level specified. If you want your client to have additional
        announced things here, then implement the class `_speak` for your
        client.

        """
        if self.quiet is False:
            bot.info("[client|%s] [database|%s]" % (self.client_name, self.database))

            if hasattr(self, "_speak"):
                self._speak()

    def _speak(self):
        """this function should be subclassed if the client has additional
        information to give the user, beyond it's name and the database
        location. Be careful about adding extra prints to various functions
        because with a command like "get" the expectation is to print a
        download url (and nothing else)
        """
        pass  # pylint: disable=unnecessary-pass

    def announce(self, command=None):
        """the client will announce itself given that a command is not in a
        particular predefined list.
        """
        if command is not None:
            if command not in ["get"] and self.quiet is False:
                self.speak()

    def __repr__(self):
        return "[client][%s]" % self.client_name

    def __str__(self):
        return "[client][%s]" % self.client_name

    def client_name(self):
        return self.__module__.split(".")[-1]


# Headers
ApiConnection._get_headers = get_headers
ApiConnection._reset_headers = reset_headers
ApiConnection._update_headers = update_headers

# Settings
ApiConnection.require_secrets = require_secrets
ApiConnection._get_setting = get_setting
ApiConnection._get_settings = get_settings
ApiConnection._get_and_update_setting = get_and_update_setting
ApiConnection._required_get_and_update = required_get_and_update
ApiConnection._get_storage_name = get_storage_name
ApiConnection._update_setting = update_setting

# Metadata
ApiConnection.get_metadata = get_metadata

# Auth
ApiConnection.require_secrets = require_secrets
ApiConnection._auth_flow = auth_flow

# Http and Requests
ApiConnection._call = call
ApiConnection._delete = delete
ApiConnection.download = download
ApiConnection._get = get
ApiConnection._head = head
ApiConnection._healthy = healthy
ApiConnection._paginate_get = paginate_get
ApiConnection._post = post
ApiConnection._put = put
ApiConnection.stream = stream
ApiConnection._stream = stream_response
ApiConnection._verify = verify



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
from sregistry.utils import parse_image_name, remove_uri

import sys
import os


def push(self, path, name, tag=None):
    """push an image to your Storage. If the collection doesn't exist,
    it is created.

    Parameters
    ==========
    path: should correspond to an absolute image path (or derive it)
    name: should be the complete uri that the user has requested to push.
    tag: should correspond with an image tag. This is provided to mirror Docker

    """
    path = os.path.abspath(path)
    bot.debug("PUSH %s" % path)

    if not os.path.exists(path):
        bot.exit("%s does not exist." % path)

    # Parse image names
    names = parse_image_name(remove_uri(name), tag=tag)

    # Get the size of the file
    file_size = os.path.getsize(path)

    # Create / get the collection
    self._get_or_create_collection(names["collection"])

    # The image name is the name followed by tag
    image_name = os.path.basename(names["storage"])

    # prepare the progress bar
    progress = 0
    bot.show_progress(progress, file_size, length=35)

    # Put the (actual) container into the collection
    with open(path, "rb") as F:
        self.conn.put_object(
            names["collection"],
            image_name,
            contents=F.read(),
            content_type="application/octet-stream",
        )

    # Finish up
    bot.show_progress(
        iteration=file_size, total=file_size, length=35, carriage_return=True
    )

    # Newline to finish download
    sys.stdout.write("\n")



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
from sregistry.utils import parse_image_name, remove_uri


def pull(self, images, file_name=None, save=True, force=False, **kwargs):
    """pull an image from a docker hub. This is a (less than ideal) workaround
       that actually does the following:

       - creates a sandbox folder
       - adds docker layers, metadata folder, and custom metadata to it
       - converts to a squashfs image with build

    the docker manifests are stored with registry metadata.

    Parameters
    ==========
    images: refers to the uri given by the user to pull in the format
    <collection>/<namespace>. You should have an API that is able to
    retrieve a container based on parsing this uri.
    file_name: the user's requested name for the file. It can
               optionally be None if the user wants a default.
    save: if True, you should save the container to the database
          using self.add()

    Returns
    =======
    finished: a single container path, or list of paths
    """

    if not isinstance(images, list):
        images = [images]

    bot.debug("Execution of PULL for %s images" % len(images))

    # If used internally we want to return a list to the user.

    finished = []
    for image in images:
        q = parse_image_name(remove_uri(image), default_collection="nvidia")

        image_file = self._pull(
            file_name=file_name,
            uri="nvidia://",
            save=save,
            force=force,
            names=q,
            kwargs=kwargs,
        )

        finished.append(image_file)

    if len(finished) == 1:
        finished = finished[0]
    return finished



"""

Copyright (C) 2017-2022 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
from sregistry.utils import read_file
import os
import globus_sdk


def parse_endpoint_name(self, endpoint):
    """split an endpoint name by colon, as the user can provide an
    endpoint name separated from a path:

    Parameters
    ==========
    endpoint 12345:/path/on/remote

    """
    parts = [x for x in endpoint.split(":") if x]
    endpoint = parts[0]
    if len(parts) == 1:
        path = ""
    else:
        path = "/".join(parts[1:])

    return endpoint, path


def create_endpoint_cache(self, endpoint_id, cache=".singularity/shub"):
    """create a directory for sregistry in the user's
    base folder to share images.

    Parameters
    ==========
    endpoint_id: the endpoint id parameters
    cache: the relative path for the images cache folder at the
           root of the endpoint

    """
    self._create_endpoint_folder(endpoint_id, cache)


def create_endpoint_folder(self, endpoint_id, folder):
    """create an endpoint folder, catching the error if it exists.

    Parameters
    ==========
    endpoint_id: the endpoint id parameters
    folder: the relative path of the folder to create

    """
    res = self.transfer_client.operation_mkdir(endpoint_id, folder)
    bot.info("%s --> %s" % (res["message"], folder))
    return res


def get_endpoint_path(self, endpoint_id):
    """return the first fullpath to a folder in the endpoint based on
    expanding the user's home from the globus config file. This
    function is fragile but I don't see any other way to do it.

    Parameters
    ==========
    endpoint_id: the endpoint id to look up the path for

    """
    config = os.path.expanduser("~/.globusonline/lta/config-paths")
    if not os.path.exists(config):
        bot.exit("%s not found for a local Globus endpoint.")

    path = None

    # Read in the config and get the root path

    config = [x.split(",")[0] for x in read_file(config)]
    for path in config:
        if os.path.exists(path):
            break

    # If we don't have an existing path, exit

    if path is None:
        bot.exit("No path was found for a local Globus endpoint.")

    return path


def init_transfer_client(self):
    """return a transfer client for the user"""

    if self._tokens_need_update():
        self._update_tokens()

    # Createe Refresh Token Authorizer

    authorizer = globus_sdk.RefreshTokenAuthorizer(
        self.transfer["refresh_token"],
        self._client,
        access_token=self.transfer["access_token"],
        expires_at=self.transfer["expires_at_seconds"],
    )

    self.transfer_client = globus_sdk.TransferClient(authorizer=authorizer)


def get_endpoint(self, endpoint_id):
    """use a transfer client to get a specific endpoint based on an endpoint id.

    Parameters
    ==========
    endpoint_id: the endpoint_id to retrieve

    """
    endpoint = None

    if not hasattr(self, "transfer_client"):
        self._init_transfer_client()

    try:
        endpoint = self.transfer_client.get_endpoint(endpoint_id).data
    except:
        bot.info("%s does not exist." % endpoint_id)

    return endpoint


def get_endpoints(self, query=None):
    """use a transfer client to get endpoints. If a search term is included,
    we use it to search a scope of "all" in addition to personal and shared
    endpoints. Endpoints are organized
    by type (my-endpoints, shared-with-me, optionally all) and then id.

    Parameters
    ==========
    query: an endpoint search term to add to a scope "all" search. If not
           defined, no searches are done with "all"

    """
    self.endpoints = {}

    if not hasattr(self, "transfer_client"):
        self._init_transfer_client()

    # We assume the user wants to always see owned and shared

    scopes = {"my-endpoints": None, "shared-with-me": None}

    # If the user provides query, add to search

    if query is not None:
        scopes.update({"all": query})

    for scope, q in scopes.items():
        self.endpoints[scope] = {}
        for ep in self.transfer_client.endpoint_search(q, filter_scope=scope):
            ep = ep.__dict__["_data"]
            self.endpoints[scope][ep["id"]] = ep

    # Alert the user not possible without personal lookup

    if len(self.endpoints["my-endpoints"]) == 0:
        bot.warning("No personal endpoint found for local transfer.")
        bot.warning("https://www.globus.org/globus-connect-personal")

    return self.endpoints



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.main import ApiConnection

from .pull import pull
from .query import search, list_all, search_collection


class Client(ApiConnection):
    def __init__(self, secrets=None, **kwargs):
        self._update_headers()
        super(Client, self).__init__(**kwargs)
        self.base = "https://www.singularity-hub.org/api"


Client.pull = pull
Client.search = search
Client.list = list_all
Client._search_collection = search_collection



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sqlalchemy import (
    create_engine,
    Column,
    DateTime,
    Integer,
    String,
    Text,
    ForeignKey,
    func,
)

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.schema import UniqueConstraint
from sqlalchemy.orm import backref, relationship, scoped_session, sessionmaker

from sregistry.logger import bot
from sregistry.defaults import SREGISTRY_STORAGE
from uuid import uuid4
import os

Base = declarative_base()


class Collection(Base):
    """A participant in a local assessment. id must be unique. If a token is
    revoked or finished, it will end with `_revoked` or `_finished`. A
    user generated without a token will have value of None
    """

    __tablename__ = "collection"
    id = Column(Integer, primary_key=True)
    name = Column(String(150), unique=True)
    token = Column(String(50))
    created_at = Column(DateTime, default=func.now())
    containers = relationship(
        "Container",
        lazy="select",
        cascade="delete,all",
        backref=backref("collection", lazy="joined"),
    )

    def __init__(self, name=None, token=None):
        self.name = name
        if token is None:
            token = str(uuid4())
        self.token = token

    def __repr__(self):
        return "<Collection %r>" % (self.name)

    def __str__(self):
        return "<Collection %r>" % (self.name)

    def url(self):
        """return the collection url"""
        return "file://"


class Container(Base):
    """a container belongs to a collection

    Parameters
    ==========
    created_at: the creation date of the image / container
    metrics: typically the inspection of the image. If not possible, then the
             basic name (uri) derived from the user is used.
    tag: the image tag
    image: the path to the image on the filesystem (can be Null)
    url: the url where the imate was ultimately retrieved, call be Null
    client: the client backend associated with the image, the type(client)
    version: a version string associated with the image
    :collection_id: the id of the colletion to which the image belongs.

    We index / filter containers based on the full uri, which is assembled
               from the <collection>/<namespace>:<tag>@<version>, then stored
               as a variable, and maintained separately for easier query.
    """

    __tablename__ = "container"
    id = Column(Integer, primary_key=True)
    created_at = Column(DateTime, default=func.now())
    metrics = Column(Text, nullable=False)
    name = Column(String(250), nullable=False)
    tag = Column(String(250), nullable=False)
    image = Column(String(250), nullable=True)
    url = Column(String(500), nullable=True)
    uri = Column(String(500), nullable=True)
    client = Column(String(50), nullable=False)
    version = Column(String(250), nullable=True)
    collection_id = Column(Integer, ForeignKey("collection.id"), nullable=False)

    __table_args__ = (
        UniqueConstraint(
            "collection_id", "name", "tag", "client", "version", name="_container_uc"
        ),
    )

    def __repr__(self):
        if self.uri is None:
            return "<Container %r>" % (self.name)
        return "<Container %r>" % (self.uri)

    def __str__(self):
        if self.uri is None:
            return "<Container %r>" % (self.name)
        return "<Container %r>" % (self.uri)

    def get_uri(self):
        """generate a uri on the fly from database parameters if one is not
        saved with the initial model (it should be, but might not be possible)
        """
        uri = "%s/%s:%s" % (self.collection.name, self.name, self.tag)
        if self.version not in [None, ""]:
            uri = "%s@%s" % (uri, self.version)
        return uri


def init_db(self, db_path):
    """initialize the database, with the default database path or custom of

    the format sqlite:////home/<username>/sregistry.db

    The custom path can be set with the environment var SREGISTRY_DATABASE
    when a user creates the client, we must initialize this db
    the database should use the .singularity cache folder to cache
    layers and images, and .singularity/sregistry.db as a database
    """

    # Database Setup, use default if uri not provided
    self.database = "sqlite:///%s" % db_path
    self.storage = SREGISTRY_STORAGE

    # If the path isn't defined, cut out early
    if not db_path:
        return

    # Ensure that the parent_folder exists)
    parent_folder = os.path.dirname(db_path)

    # Case 1: Does not exist
    if not os.path.exists(parent_folder):
        bot.exit("Database location {} does not exist.".format(parent_folder))

    # Case 2: Insufficient permission for write
    if not os.access(parent_folder, os.W_OK):
        bot.exit("Insufficient permission to write to {}".format(parent_folder))

    bot.debug("Database located at %s" % self.database)
    self.engine = create_engine(self.database, convert_unicode=True)
    self.session = scoped_session(
        sessionmaker(autocommit=False, autoflush=False, bind=self.engine)
    )

    Base.query = self.session.query_property()

    # import all modules here that might define models so that
    # they will be registered properly on the metadata.  Otherwise
    # you will have to import them first before calling init_db()
    Base.metadata.create_all(bind=self.engine)
    self.Base = Base



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.utils import parse_image_name, remove_uri, which

from sregistry.logger import bot
import os
import json

# Metadata


def get_metadata(self, image_file, names=None):
    """extract metadata using Singularity inspect, if the executable is found.
    If not, return a reasonable default (the parsed image name)

    Parameters
    ==========
    image_file: the full path to a Singularity image
    names: optional, an extracted or otherwise created dictionary of
           variables for the image, likely from utils.parse_image_name

    """
    if names is None:
        names = {}

    metadata = {}

    # We can't return anything without image_file or names
    if image_file:
        if not os.path.exists(image_file):
            bot.error("Cannot find %s." % image_file)
            return names or metadata

    # The user provided a file, but no names
    if not names:
        names = parse_image_name(remove_uri(image_file))

    # Look for the Singularity Executable
    singularity = which("singularity")["message"]

    # Inspect the image, or return names only
    if os.path.exists(singularity) and image_file:
        from spython.main import Client as Singularity

        # Store the original quiet setting
        is_quiet = Singularity.quiet

        # We try and inspect, but not required (wont work within Docker)
        try:
            Singularity.quiet = True
            updates = Singularity.inspect(image=image_file)
        except:
            bot.warning("Inspect command not supported, metadata not included.")
            updates = None

        # Restore the original quiet setting
        Singularity.quiet = is_quiet

        # Try loading the metadata
        if updates is not None:
            try:
                updates = json.loads(updates)

                # Singularity 3.x bug with missing top level
                if "data" in updates:
                    updates = updates["data"]
                metadata.update(updates)

                # Flatten labels
                if "attributes" in metadata:
                    if "labels" in metadata["attributes"]:
                        metadata.update(metadata["attributes"]["labels"])
                    del metadata["attributes"]

            except:
                pass

    metadata.update(names)

    # Add the type to the container
    metadata["type"] = "container"

    return metadata



"""

pull.py: pull function for singularity registry

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from requests.exceptions import SSLError
from requests.models import Response
from sregistry.utils import parse_image_name, remove_uri
from sregistry.logger import bot
import os


def pull(self, images, file_name=None, save=True, **kwargs):
    """pull an image from a singularity registry

    Parameters
    ==========
    images: refers to the uri given by the user to pull in the format
    <collection>/<namespace>. You should have an API that is able to
    retrieve a container based on parsing this uri.
    file_name: the user's requested name for the file. It can
               optionally be None if the user wants a default.
    save: if True, you should save the container to the database
          using self.add()

    Returns
    =======
    finished: a single container path, or list of paths
    """

    if not isinstance(images, list):
        images = [images]

    bot.debug("Execution of PULL for %s images" % len(images))

    finished = []
    for image in images:
        q = parse_image_name(remove_uri(image))

        # If a custom registry is not set, use default base
        if q["registry"] is None:
            q["registry"] = self.base

        # If the registry is still None, no go
        if q["registry"] is None:
            bot.exit("You must define a base in secrets, image uri, or environment")

        # Ensure https is added back to the registry  uri
        q = self._add_https(q)

        # All custom registries need api appended
        if not q["registry"].endswith("api"):
            q["registry"] = "%s/api" % q["registry"]

        # Verify image existence, and obtain id
        url = "%s/container/%s/%s:%s" % (
            q["registry"],
            q["collection"],
            q["image"],
            q["tag"],
        )

        bot.debug("Retrieving manifest at %s" % url)

        try:
            manifest = self._get(url)
        except SSLError:
            bot.exit("Issue with %s, try exporting SREGISTRY_REGISTRY_NOHTTPS." % url)

        # Private container collection
        if isinstance(manifest, Response):
            # Requires token
            if manifest.status_code in [403, 401]:
                SREGISTRY_EVENT = self.authorize(request_type="pull", names=q)
                headers = {"Authorization": SREGISTRY_EVENT}
                self._update_headers(headers)
                manifest = self._get(url)

                # Still denied
                if isinstance(manifest, Response):
                    if manifest.status_code == 403:
                        manifest = 403

        if isinstance(manifest, int):
            if manifest == 400:
                bot.exit("Bad request (400). Is this a private container?")
            elif manifest == 404:
                bot.exit("Container not found (404)")
            elif manifest == 403:
                bot.exit("Unauthorized (403)")

        # Successful pull
        if "image" in manifest:
            # Add self link to manifest
            manifest["selfLink"] = url

            if file_name is None:
                file_name = q["storage"].replace("/", "-")

            # Clear headers of previous token
            self._reset_headers()

            # Show progress if not quiet
            image_file = self.download(
                url=manifest["image"], file_name=file_name, show_progress=not self.quiet
            )

            # If the user is saving to local storage
            if save is True:
                image_uri = "%s/%s:%s" % (
                    manifest["collection"],
                    manifest["name"],
                    manifest["tag"],
                )
                container = self.add(
                    image_path=image_file,
                    image_uri=image_uri,
                    metadata=manifest,
                    url=manifest["image"],
                )
                image_file = container.image

            if os.path.exists(image_file):
                bot.debug("Retrieved image file %s" % image_file)
                bot.custom(prefix="Success!", message=image_file)
                finished.append(image_file)

    if len(finished) == 1:
        finished = finished[0]
    return finished



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import sys

import time
import threading
from random import choice


class Spinner:
    spinning = False
    delay = 0.1

    @staticmethod
    def spinning_cursor():
        while 1:
            for cursor in "|/-\\":
                yield cursor

    @staticmethod
    def balloons_cursor():
        while 1:
            for cursor in ". o O @ *":
                yield cursor

    @staticmethod
    def changing_arrows():
        while 1:
            for cursor in "<^>v":
                yield cursor

    def select_generator(self, generator):
        if generator is None:
            generator = choice(["cursor", "arrow", "balloons"])

        return generator

    def __init__(self, delay=None, generator=None):
        generator = self.select_generator(generator)

        if generator == "cursor":
            self.spinner_generator = self.spinning_cursor()
        elif generator == "arrow":
            self.spinner_generator = self.changing_arrows()
        elif generator == "balloons":
            self.spinner_generator = self.balloons_cursor()
            if delay is None:
                delay = 0.2
        else:
            self.spinner_generator = self.spinning_cursor()

        if delay and float(delay):
            self.delay = delay

    def run(self):
        while self.spinning:
            sys.stdout.write(next(self.spinner_generator))
            sys.stdout.flush()
            time.sleep(self.delay)
            sys.stdout.write("\b")
            sys.stdout.flush()

    def start(self):
        self.spinning = True
        threading.Thread(target=self.run).start()

    def stop(self):
        self.spinning = False
        time.sleep(self.delay)



"""

auth.py: authorization functions for client

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.main.registry.utils import (
    generate_signature,
    generate_credential,
    generate_timestamp,
)

from sregistry.logger import bot


def authorize(self, names, payload=None, request_type="push"):
    """Authorize a client based on encrypting the payload with the client
    token, which should be matched on the receiving server"""

    self.username = self._get_and_update_setting("SREGISTRY_REGISTRY_USERNAME")

    if self.username is None:
        # backwards compatibility
        self.username = self._get_and_update_setting("username")

        if self.username is None:
            bot.exit(
                "Failed to authorize: please set SREGISTRY_REGISTRY_USERNAME to an appropriate value"
            )
        else:
            self._update_setting("SREGISTRY_REGISTRY_USERNAME", self.username)

    self.token = self._get_and_update_setting("SREGISTRY_REGISTRY_TOKEN")

    if self.token is None:
        # backwards compatibility
        self.token = self._get_and_update_setting("token")

        if self.token is None:
            bot.exit(
                "Failed to authorize: please set SREGISTRY_REGISTRY_TOKEN to an appropriate value"
            )
        else:
            self._update_setting("SREGISTRY_REGISTRY_TOKEN", self.token)

    # Use the payload to generate a digest   push|collection|name|tag|user
    timestamp = generate_timestamp()
    credential = generate_credential(self.username)
    credential = "%s/%s/%s" % (request_type, credential, timestamp)

    if payload is None:
        payload = "%s|%s|%s|%s|%s|" % (
            request_type,
            names["collection"],
            timestamp,
            names["image"],
            names["tag"],
        )

    signature = generate_signature(payload, self.token)
    return "SREGISTRY-HMAC-SHA256 Credential=%s,Signature=%s" % (credential, signature)



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
import sys


def share(self, query, share_to):
    """share will use the client to get an image based on a query, and then
    the link with an email or endpoint (share_to) of choice.
    """

    images = self._container_query(query, quiet=True)
    if len(images) == 0:
        bot.info("Cannot find a remote image matching %s" % query)
        sys.exit(0)

    image = images[0]

    def callback(request_id, response, exception):
        if exception:
            # Handle error
            print(exception)
        else:
            share_id = response.get("id")
            bot.info("Share to %s complete: %s!" % (share_to, share_id))

    batch = self._service.new_batch_http_request(callback=callback)
    user_permission = {"type": "user", "role": "reader", "emailAddress": share_to}

    batch.add(
        self._service.permissions().create(
            fileId=image["id"], body=user_permission, fields="id"
        )
    )

    batch.execute()
    return image



from .terminal import (
    run_command,
    check_install,
    get_installdir,
    get_thumbnail,
    which,
    confirm_action,
    confirm_delete,
)
from .names import get_recipe_tag, get_uri, parse_image_name, remove_uri
from .fileio import (
    copyfile,
    extract_tar,
    get_userhome,
    get_file_hash,
    get_tmpdir,
    get_tmpfile,
    mkdir_p,
    print_json,
    read_file,
    read_json,
    write_file,
    write_json,
)



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
from sregistry.defaults import SREGISTRY_STORAGE
from sregistry.utils import parse_image_name
import globus_sdk
import os


def push(self, path, name, tag=None):
    """push an image to Globus endpoint. In this case, the name is the
    globus endpoint id and path.

    --name <endpointid>:/path/for/image

    """

    # Split the name into endpoint and rest

    endpoint, _ = self._parse_endpoint_name(name)

    path = os.path.abspath(path)
    image = os.path.basename(path)
    bot.debug("PUSH %s" % path)

    # Flatten image uri into image name

    q = parse_image_name(image)

    if not os.path.exists(path):
        bot.exit("%s does not exist." % path)

    # Ensure we have a transfer client
    if not hasattr(self, "transfer_client"):
        self._init_transfer_client()

    # The user must have a personal endpoint

    endpoints = self._get_endpoints()

    if len(endpoints["my-endpoints"]) == 0:
        bot.exit("You must have a personal endpoint to transfer the container")

    # Take the first endpoint that is active

    source_endpoint = None
    for _, contender in endpoints["my-endpoints"].items():
        if contender["gcp_connected"] is True:
            source_endpoint = contender
            break

    # Exit if none are active, required!
    if source_endpoint is None:
        bot.exit("No activated local endpoints online! Go online to transfer")

    # The destination endpoint should have an .singularity/shub folder set
    self._create_endpoint_cache(endpoint)

    # SREGISTRY_STORAGE must be an endpoint
    # if the image isn't already there, add it first

    added = self.add(image_path=path, image_uri=q["uri"], copy=True)

    label = "Singularity Registry Transfer for %s" % added.name
    tdata = globus_sdk.TransferData(
        self.transfer_client,
        source_endpoint["id"],
        endpoint,
        label=label,
        sync_level="checksum",
    )
    image = ".singularity/shub/%s" % image
    tdata.add_item(added.image, image)
    bot.info(
        "Requesting transfer from local %s to %s:%s"
        % (SREGISTRY_STORAGE, endpoint, image)
    )
    transfer_result = self.transfer_client.submit_transfer(tdata)
    bot.info(transfer_result["message"])
    return transfer_result



"""

sregistry.api: base template for making a connection to an API

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from requests.exceptions import HTTPError
import requests

from sregistry.utils import get_tmpfile
from sregistry.logger import bot
import shutil
import json
import sys


def delete(self, url, headers=None, return_json=True, default_headers=True):
    """delete request, use with caution"""
    bot.debug("DELETE %s" % url)
    return self._call(
        url,
        headers=headers,
        func=requests.delete,
        return_json=return_json,
        default_headers=default_headers,
    )


def head(self, url):
    """head request, typically used for status code retrieval, etc."""
    bot.debug("HEAD %s" % url)
    return self._call(url, func=requests.head)


def healthy(self, url):
    """determine if a resource is healthy based on an accepted response (200)
    or redirect (301)

    Parameters
    ==========
    url: the URL to check status for, based on the status_code of HEAD

    """
    response = requests.get(url)
    status_code = response.status_code
    if status_code != 200:
        bot.error("%s, response status code %s." % (url, status_code))
        return False
    return True


def put(self, url, headers=None, data=None, return_json=True, default_headers=True):
    """put request"""
    bot.debug("PUT %s" % url)
    return self._call(
        url,
        headers=headers,
        func=requests.put,
        data=data,
        return_json=return_json,
        default_headers=default_headers,
    )


def post(self, url, headers=None, data=None, return_json=True, default_headers=True):
    """post will use requests to get a particular url"""

    bot.debug("POST %s" % url)
    return self._call(
        url,
        headers=headers,
        func=requests.post,
        data=data,
        return_json=return_json,
        default_headers=default_headers,
    )


def get(
    self,
    url,
    headers=None,
    token=None,
    data=None,
    return_json=True,
    default_headers=True,
    quiet=False,
):
    """get will use requests to get a particular url"""
    bot.debug("GET %s" % url)
    return self._call(
        url,
        headers=headers,
        func=requests.get,
        data=data,
        return_json=return_json,
        default_headers=default_headers,
        quiet=quiet,
    )


def paginate_get(self, url, headers=None, return_json=True, start_page=None):
    """paginate_call is a wrapper for get to paginate results"""

    geturl = "%s&page=1" % (url)
    if start_page is not None:
        geturl = "%s&page=%s" % (url, start_page)

    results = []
    while geturl is not None:
        result = self._get(url, headers=headers, return_json=return_json)
        # If we have pagination:
        if isinstance(result, dict):
            if "results" in result:
                results = results + result["results"]
            geturl = result["next"]
        # No pagination is a list
        else:
            return result
    return results


def verify(self):
    """
    verify will return a True or False to determine to verify the
    requests call or not. If False, we should the user a warning message,
    as this should not be done in production!

    """
    from sregistry.defaults import DISABLE_SSL_CHECK

    if DISABLE_SSL_CHECK is True:
        bot.warning("Verify of certificates disabled! ::TESTING USE ONLY::")

    return not DISABLE_SSL_CHECK


def download(self, url, file_name, headers=None, show_progress=True):
    """stream to a temporary file, rename on successful completion

    Parameters
    ==========
    file_name: the file name to stream to
    url: the url to stream from
    headers: additional headers to add
    force: If the final image exists, don't overwrite
    show_progress: boolean to show progress bar
    """

    tmp_file = get_tmpfile(prefix="%s.tmp." % file_name)

    # Should we verify the request?
    verify = self._verify()

    # Check here if exists
    if requests.head(url, verify=verify).status_code in [200, 401]:
        response = self.stream(
            url, headers=headers, stream_to=tmp_file, show_progress=show_progress
        )

        if isinstance(response, HTTPError):
            bot.exit("Error downloading %s, exiting." % url)

        shutil.move(tmp_file, file_name)
    else:
        bot.error("Invalid url or permissions %s" % url)
    return file_name


def stream(
    self,
    url,
    headers=None,
    stream_to=None,
    retry=True,
    default_headers=True,
    show_progress=True,
):
    """
    stream is a get that will stream to file_name. This stream is intended
    to take a url and (optionally) a set of headers and file to stream to,
    and will generate a response with requests.get.

    Parameters
    ==========
    url: the url to do a requests.get to
    headers: any updated headers to use for the requets
    stream_to: the file to stream to
    show_progress: boolean to show progress bar
    retry: should the client retry? (intended for use after token refresh)
           by default we retry once after token refresh, then fail.
    """
    bot.debug("GET %s" % url)

    # Ensure headers are present, update if not
    if headers is None:
        if self.headers is None:
            self._reset_headers()
        headers = self.headers.copy()

    response = requests.get(url, headers=headers, verify=self._verify(), stream=True)

    # Deal with token if necessary
    if response.status_code == 401 and retry is True:
        if hasattr(self, "_update_token"):
            self._update_token(response)
            return self.stream(
                url, headers, stream_to, retry=False, show_progress=show_progress
            )

    if response.status_code == 200:
        return self._stream(response, stream_to=stream_to, show_progress=show_progress)

    bot.exit("Problem with stream, response %s" % (response.status_code))


def stream_response(self, response, stream_to=None, show_progress=True):
    """
    stream response is one level higher up than stream, starting with a
    response object and then performing the stream without making the
    requests.get. The expectation is that the request was successful
    (status code 20*).
    show_progress: boolean to show progress bar

    Parameters
    ==========
    response: a response that is ready to be iterated over to download in
              streamed chunks
    stream_to: the file to stream to


    """
    if response.status_code == 200:
        if show_progress is False:
            bot.quiet = True

        # Keep user updated with Progress Bar, if not quiet
        content_size = None
        if "Content-Length" in response.headers:
            progress = 0
            content_size = int(response.headers["Content-Length"])
            bot.show_progress(progress, content_size, length=35)

        chunk_size = 1 << 20
        with open(stream_to, "wb") as filey:
            for chunk in response.iter_content(chunk_size=chunk_size):
                filey.write(chunk)
                if content_size is not None:
                    progress += chunk_size
                    bot.show_progress(
                        iteration=progress,
                        total=content_size,
                        length=35,
                        carriage_return=False,
                    )

        # Newline to finish download
        sys.stdout.write("\n")

        return stream_to

    bot.exit("Problem with stream, response %s" % (response.status_code))


def call(
    self,
    url,
    func,
    data=None,
    headers=None,
    return_json=True,
    stream=False,
    retry=True,
    default_headers=True,
    quiet=False,
):
    """call will issue the call, and issue a refresh token
    given a 401 response, and if the client has a _update_token function

    Parameters
    ==========
    func: the function (eg, post, get) to call
    url: the url to send file to
    headers: if not None, update the client self.headers with dictionary
    data: additional data to add to the request
    return_json: return json if successful
    default_headers: use the client's self.headers (default True)

    """

    if data is not None:
        if not isinstance(data, dict):
            data = json.dumps(data)

    heads = dict()
    if default_headers is True:
        heads = self.headers.copy()

    if headers is not None:
        if isinstance(headers, dict):
            heads.update(headers)

    response = func(
        url=url, headers=heads, data=data, verify=self._verify(), stream=stream
    )

    # Errored response, try again with refresh
    if response.status_code in [500, 502]:
        bot.exit("Beep boop! %s: %s" % (response.reason, response.status_code))

    # Errored response, try again with refresh
    if response.status_code == 404:
        # Not found, we might want to continue on
        if quiet is False:
            bot.exit("Beep boop! %s: %s" % (response.reason, response.status_code))

    # Errored response, try again with refresh
    if response.status_code == 401:
        # If client has method to update token, try it once
        if retry is True and hasattr(self, "_update_token"):
            # A result of None indicates no update to the call
            self._update_token(response)
            return self._call(
                url,
                func,
                data=data,
                headers=headers,
                return_json=return_json,
                stream=stream,
                retry=False,
            )

        bot.exit(
            "Your credentials are expired! %s: %s"
            % (response.reason, response.status_code)
        )

    elif response.status_code == 200:
        if return_json:
            try:
                response = response.json()
            except ValueError:
                bot.exit("The server returned a malformed response.")

    return response



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot, ProgressBar
from sregistry.utils import parse_image_name, remove_uri
from googleapiclient.http import MediaIoBaseDownload
import os


def pull(self, images, file_name=None, save=True, **kwargs):
    """pull an image from google drive, based on a query (uri or id)

    Parameters
    ==========
    images: refers to the uri given by the user to pull in the format
    <collection>/<namespace>. You should have an API that is able to
    retrieve a container based on parsing this uri.
    file_name: the user's requested name for the file. It can
               optionally be None if the user wants a default.
    save: if True, you should save the container to the database
          using self.add()

    Returns
    =======
    finished: a single container path, or list of paths
    """

    if not isinstance(images, list):
        images = [images]

    bot.debug("Execution of PULL for %s images" % len(images))

    # If used internally we want to return a list to the user.
    finished = []
    for image in images:
        q = parse_image_name(remove_uri(image))

        # Use container search to find the container based on uri
        bot.info("Searching for %s in drive://%s" % (q["uri"], self._base))
        matches = self._container_query(q["uri"], quiet=True)

        if len(matches) == 0:
            bot.exit("No matching containers found.")

        # If the user didn't provide a file, make one based on the names
        if file_name is None:
            file_name = q["storage"].replace("/", "-")

        # We give the first match, the uri should be unique and known
        image = matches[0]

        request = self._service.files().get_media(fileId=image["id"])

        with open(file_name, "wb") as fh:
            downloader = MediaIoBaseDownload(fh, request)
            done = False
            bar = None

            # Download and update the user with progress bar
            while done is False:
                status, done = downloader.next_chunk()

                # Create bar on first call
                if bar is None:
                    total = status.total_size / (1024 * 1024.0)
                    bar = ProgressBar(
                        expected_size=total, filled_char="=", hide=self.quiet
                    )

                bar.show(status.resumable_progress / (1024 * 1024.0))

        # If the user is saving to local storage, you need to assumble the uri
        # here in the expected format <collection>/<namespace>:<tag>@<version>
        if save is True:
            image_uri = q["uri"]
            if "uri" in image:
                image_uri = image["uri"]

            # Update metadata with selfLink
            image["selfLink"] = downloader._uri

            container = self.add(
                image_path=file_name,
                image_uri=image_uri,
                metadata=image,
                url=downloader._uri,
            )

            # When the container is created, this is the path to the image
            image_file = container.image

        if os.path.exists(image_file):
            bot.debug("Retrieved image file %s" % image_file)
            bot.custom(prefix="Success!", message=image_file)
            finished.append(image_file)

    if len(finished) == 1:
        finished = finished[0]
    return finished



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.auth import read_client_secrets
from sregistry.main import ApiConnection
import json
import os

from .auth import authorize
from .build import build
from .pull import pull
from .push import push
from .delete import delete
from .query import search, search_all, collection_search, label_search, container_search


class Client(ApiConnection):
    def __init__(self, secrets=None, base=None, **kwargs):
        self.base = base
        super(Client, self).__init__(**kwargs)
        self._update_secrets()
        self._update_headers()

    def _update_base(self):
        if self.base is not None:
            self.base = self.base.strip("/")
            if not self.base.endswith("api"):
                self.base = "%s/api" % self.base

    def _read_response(self, response, field="detail"):
        """attempt to read the detail provided by the response. If none,
        default to using the reason"""

        try:
            message = json.loads(response._content.decode("utf-8"))[field]
        except:
            message = response.reason
        return message

    def _add_https(self, q):
        """for push, pull, and other api interactions, the user can optionally
        define a custom registry. If the registry name doesn't include http
        or https, add it.

        Parameters
        ==========
        q: the parsed image query (names), including the original
        """

        # If image uses http or https, add back
        if not q["registry"].startswith("http"):
            if q["original"].startswith("http:"):
                q["registry"] = "http://%s" % q["registry"]

            elif q["original"].startswith("https:"):
                q["registry"] = "https://%s" % q["registry"]

            # Otherwise, guess from the user's environment
            else:
                prefix = "https://"

                # The user can set an environment variable to specify nohttps
                nohttps = os.environ.get("SREGISTRY_REGISTRY_NOHTTPS")
                if nohttps is not None:
                    prefix = "http://"
                q["registry"] = "%s%s" % (prefix, q["registry"])

        return q

    def _update_secrets(self):
        """update secrets will use the SREGISTRY_REGISTRY_BASE to determine which API base url
        to use for the registry. If not present, the .sregistry secrets credential file
        either located at $HOME/.sregistry or the environment variable
        SREGISTRY_CLIENT_SECRETS will be used instead.
        """
        self.base = self._get_and_update_setting("SREGISTRY_REGISTRY_BASE")

        if self.base is None:
            self.base = self._get_and_update_setting("base")

            if self.base is not None:
                self._update_setting("SREGISTRY_REGISTRY_BASE", self.base)

        self._update_base()

    def __str__(self):
        return type(self)


Client.authorize = authorize
Client.build = build
Client.delete = delete
Client.pull = pull
Client.push = push
Client.search = search
Client._search_all = search_all
Client._collection_search = collection_search
Client._container_search = container_search
Client._label_search = label_search



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot, ProgressBar
from sregistry.utils import get_file_hash, parse_image_name, remove_uri

from googleapiclient.http import MediaFileUpload
from retrying import retry
import os


def push(self, path, name, tag=None):
    """push an image to Google Cloud Storage, meaning uploading it

    Parameters
    ==========
    path: should correspond to an absolte image path (or derive it)
    name: should be the complete uri that the user has requested to push.
    tag: should correspond with an image tag. This is provided to mirror Docker
    """
    path = os.path.abspath(path)
    bot.debug("PUSH %s" % path)

    if not os.path.exists(path):
        bot.exit("%s does not exist." % path)

    # This returns a data structure with collection, container, based on uri
    names = parse_image_name(remove_uri(name), tag=tag)

    if names["version"] is None:
        version = get_file_hash(path, "sha256")
        names = parse_image_name(remove_uri(name), tag=tag, version=version)

    # Update metadata with names
    metadata = self.get_metadata(path, names=names)

    manifest = self._upload(
        source=path, destination=names["storage"], metadata=metadata
    )

    print(manifest["mediaLink"])


@retry(wait_exponential_multiplier=1000, wait_exponential_max=10000)
def upload(self, source, destination, chunk_size=2 * 1024 * 1024, metadata=None):
    """upload a file from a source to a destination. The client is expected
    to have a bucket (self._bucket) that is created when instantiated.

    This would be the method to do the same using the storage client,
    but not easily done for resumable

    blob = self._bucket.blob(destination)
    blob.upload_from_filename(filename=source,
                              content_type="application/zip",
                              client=self._service)

    url = blob.public_url
    if isinstance(url, six.binary_type):
        url = url.decode('utf-8')

    return url
    """
    env = "SREGISTRY_GOOGLE_STORAGE_PRIVATE"
    keep_private = self._get_and_update_setting(env)

    media = MediaFileUpload(source, chunksize=chunk_size, resumable=True)
    request = self._storage_service.objects().insert(
        bucket=self._bucket_name, name=destination, media_body=media
    )

    response = None
    total = request.resumable._size / (1024 * 1024.0)

    bar = ProgressBar(expected_size=total, filled_char="=", hide=self.quiet)

    while response is None:
        progress, response = request.next_chunk()
        if progress:
            bar.show(progress.resumable_progress / (1024 * 1024.0))

    # When we finish upload, get as blob
    blob = self._bucket.blob(destination)
    if blob.exists():
        if not keep_private:
            blob.make_public()

        # If the user has a dictionary of metadata to update
        if metadata is not None:
            blob.metadata = metadata
            blob._properties["metadata"] = metadata
            blob.patch()

    return response



"""

Copyright (C) 2016-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.defaults import SINGULARITY_CACHE
from sregistry.logger import bot
from sregistry.utils import get_tmpdir, mkdir_p, print_json
from .utils import get_template, create_tar
import json
import math
import os
import re
import shutil
import tempfile


###############################################################################


def update_token(self, response):
    """update_token uses HTTP basic authentication to get a token for
    Docker registry API V2 operations. We get here if a 401 is
    returned for a request.

    Parameters
    ==========
    response: the http request response to parse for the challenge.

    https://docs.docker.com/registry/spec/auth/token/
    """

    not_asking_auth = "Www-Authenticate" not in response.headers
    if response.status_code != 401 or not_asking_auth:
        bot.exit("Authentication error, exiting.")

    challenge = response.headers["Www-Authenticate"]
    regexp = '^Bearer\s+realm="(.+)",service="(.+)",scope="(.+)",?'
    match = re.match(regexp, challenge)

    if not match:
        bot.exit("Unrecognized authentication challenge, exiting.")

    realm = match.group(1)
    service = match.group(2)
    scope = match.group(3).split(",")[0]
    token_url = realm + "?service=" + service + "&expires_in=900&scope=" + scope

    # Default headers must be False so that client's current headers not used
    response = self._get(token_url)

    try:
        token = response["token"]
        token = {"Authorization": "Bearer %s" % token}
        self.headers.update(token)

    except:
        bot.exit("Error getting token.")


def get_manifests(self, repo_name, digest=None):
    """get_manifests calls get_manifest for each of the schema versions,
    including v2 and v1. Version 1 includes image layers and metadata,
    and version 2 must be parsed for a specific manifest, and the 2nd
    call includes the layers. If a digest is not provided
    latest is used.

    Parameters
    ==========
    repo_name: reference to the <username>/<repository>:<tag> to obtain
    digest: a tag or shasum version

    """

    if not hasattr(self, "manifests"):
        self.manifests = {}

    # Obtain schema version 1 (metadata) and 2, and image config
    schemaVersions = ["v1", "v2", "config"]
    for schemaVersion in schemaVersions:
        manifest = self._get_manifest(repo_name, digest, schemaVersion)
        if manifest is not None:
            # If we don't have a config yet, try to get from version 2 manifest
            if schemaVersion == "v2" and "config" in manifest:
                bot.debug("Attempting to get config as blob in verison 2 manifest")
                url = self._get_layerLink(repo_name, manifest["config"]["digest"])
                headers = {"Accept": manifest["config"]["mediaType"]}
                self.manifests["config"] = self._get(url, headers=headers)

            self.manifests[schemaVersion] = manifest

    return self.manifests


def get_manifest_selfLink(self, repo_name, digest=None):
    """get a selfLink for the manifest, for use by the client get_manifest
     function, along with the parents pull

    Parameters
    ==========
    repo_name: reference to the <username>/<repository>:<tag> to obtain
    digest: a tag or shasum version

    """
    url = "%s/%s/manifests" % (self.base, repo_name)

    # Add a digest - a tag or hash (version)
    if digest is None:
        digest = "latest"
    return "%s/%s" % (url, digest)


def get_manifest(self, repo_name, digest=None, version="v1"):
    """
    get_manifest should return an image manifest
    for a particular repo and tag.  The image details
    are extracted when the client is generated.

    Parameters
    ==========
    repo_name: reference to the <username>/<repository>:<tag> to obtain
    digest: a tag or shasum version
    version: one of v1, v2, and config (for image config)

    """

    accepts = {
        "config": "application/vnd.docker.container.image.v1+json",
        "v1": "application/vnd.docker.distribution.manifest.v1+json",
        "v2": "application/vnd.docker.distribution.manifest.v2+json",
    }

    url = self._get_manifest_selfLink(repo_name, digest)

    bot.verbose("Obtaining manifest: %s %s" % (url, version))
    headers = {"Accept": accepts[version]}

    try:
        manifest = self._get(url, headers=headers, quiet=True)
        manifest["selfLink"] = url
    except:
        manifest = None

    return manifest


def download_layers(self, repo_name, digest=None, destination=None):
    """download layers is a wrapper to do the following for a client loaded
    with a manifest for an image:

    1. use the manifests to retrieve list of digests (get_digests)
    2. atomically download the list to destination (get_layers)

    This function uses the MultiProcess client to download layers
    at the same time.
    """
    from sregistry.main.workers import Workers, download_task

    # 1. Get manifests if not retrieved
    if not hasattr(self, "manifests"):
        self._get_manifests(repo_name, digest)

    # Obtain list of digets, and destination for download
    digests = self._get_digests()
    destination = self._get_download_cache(destination)

    # Create multiprocess download client
    workers = Workers()

    # Download each layer atomically
    tasks = []
    layers = []
    for digest in digests:
        targz = "%s/%s.tar.gz" % (destination, digest)

        # Only download if not in cache already
        if not os.path.exists(targz):
            url = "%s/%s/blobs/%s" % (self.base, repo_name, digest)
            tasks.append((url, self.headers, targz))
        layers.append(targz)

    # Download layers with multiprocess workers
    if len(tasks) > 0:
        workers.run(func=download_task, tasks=tasks)

    # Create the metadata tar
    metadata = self._create_metadata_tar(destination)
    if metadata is not None:
        layers.append(metadata)

    return layers


def get_download_cache(self, destination, subfolder="docker"):
    """determine the user preference for atomic download of layers. If
    the user has set a singularity cache directory, honor it. Otherwise,
    use the Singularity default.
    """
    # First priority after user specification is Singularity Cache
    if destination is None:
        destination = self._get_setting("SINGULARITY_CACHEDIR", SINGULARITY_CACHE)

        # If not set, the user has disabled (use tmp)
        destination = get_tmpdir(destination)

    if not destination.endswith(subfolder):
        destination = "%s/%s" % (destination, subfolder)

    # Create subfolders, if don't exist
    mkdir_p(destination)
    return destination


def get_digests(self):
    """return a list of layers from a manifest.
    The function is intended to work with both version
    1 and 2 of the schema. All layers (including redundant)
    are returned. By default, we try version 2 first,
    then fall back to version 1.

    For version 1 manifests: extraction is reversed

    Parameters
    ==========
    manifest: the manifest to read_layers from

    """
    if not hasattr(self, "manifests"):
        bot.exit("Please retrieve manifests for an image first.")

    digests = []

    reverseLayers = False
    schemaVersions = list(self.manifests.keys())
    schemaVersions.reverse()
    manifest = None

    # Select the manifest to use
    for schemaVersion in schemaVersions:
        manifest = self.manifests[schemaVersion]

        if manifest["schemaVersion"] == 1:
            reverseLayers = True

        # version 2 indices used by default
        layer_key = "layers"
        digest_key = "digest"

        # Docker manifest-v2-2.md#image-manifest
        if "layers" in manifest:
            bot.debug("Image manifest version 2.2 found.")
            break

        # Docker manifest-v2-1.md#example-manifest  # noqa
        elif "fsLayers" in manifest:
            layer_key = "fsLayers"
            digest_key = "blobSum"
            bot.debug("Image manifest version 2.1 found.")
            break

        else:
            msg = "Improperly formed manifest, "
            msg += "layers, manifests, or fsLayers must be present"
            bot.exit(msg)

    # If all attempts failed
    if manifest is None:
        bot.exit("Failed to find manifest. Check image name and login.")

    for layer in manifest[layer_key]:
        if digest_key in layer:
            bot.debug("Adding digest %s" % layer[digest_key])
            digests.append(layer[digest_key])

    # Reverse layer order for manifest version 1.0
    if reverseLayers is True:
        message = "reversing layers..."
        bot.debug(message)
        digests.reverse()

    return digests


def get_layerLink(self, repo_name, digest):
    """get the url for a layer based on a digest and repo name

    Parameters
    ==========
    digest: The image digest to obtain
    repo_name: the image name (library/ubuntu) to retrieve

    """
    return "%s/%s/blobs/%s" % (self.base, repo_name, digest)


def get_layer(self, image_id, repo_name, download_folder=None):
    """download an image layer (.tar.gz) to a specified download folder.

    Parameters
    ==========
    download_folder: download to this folder. If not set, uses temp.
    repo_name: the image name (library/ubuntu) to retrieve

    """
    url = self._get_layerLink(repo_name, image_id)

    bot.verbose("Downloading layers from %s" % url)

    download_folder = get_tmpdir(download_folder)
    download_folder = "%s/%s.tar.gz" % (download_folder, image_id)

    # Update user what we are doing
    bot.debug("Downloading layer %s" % image_id)

    # Step 1: Download the layer atomically
    file_name = "%s.%s" % (download_folder, next(tempfile._get_candidate_names()))

    tar_download = self.download(url, file_name)

    try:
        shutil.move(tar_download, download_folder)
    except:
        msg = "Cannot untar layer %s," % tar_download
        msg += " was there a problem with download?"
        bot.exit(msg)
    return download_folder


def get_size(self, add_padding=True, round_up=True, return_mb=True):
    """get_size will return the image size (must use v.2.0 manifest)

    Parameters
    ==========
    add_padding: if true, return reported size * 5
    round_up: if true, round up to nearest integer
    return_mb: if true, defaults bytes are converted to MB

    """
    if not hasattr(self, "manifests"):
        bot.exit("Please retrieve manifests for an image first.")

    size = 768  # default size
    for _, manifest in self.manifests.items():
        if "layers" in manifest:
            size = 0
            for layer in manifest["layers"]:
                if "size" in layer:
                    size += layer["size"]

            if add_padding is True:
                size = size * 5

            if return_mb is True:
                size = size / (1024 * 1024)  # 1MB = 1024*1024 bytes

            if round_up is True:
                size = math.ceil(size)
            size = int(size)

        return size


################################################################################
# Metadata (Environment, Labels, Runscript)
################################################################################


def get_config(self, key="Entrypoint", delim=None):
    """get_config returns a particular key (default is Entrypoint)
    from a VERSION 1 manifest obtained with get_manifest.

    Parameters
    ==========
    key: the key to return from the manifest config
    delim: Given a list, the delim to use to join the entries.
    Default is newline

    """
    if not hasattr(self, "manifests"):
        bot.exit("Please retrieve manifests for an image first.")

    cmd = None

    # If we didn't find the config value in version 2
    if cmd is None and "config" in self.manifests:
        # First try, version 2.0 manifest config has upper level config
        manifest = self.manifests["config"]
        if "config" in manifest:
            if key in manifest["config"]:
                cmd = manifest["config"][key]

        # Second try, config manifest (not from verison 2.0 schema blob)

        if cmd is None and "history" in manifest:
            for entry in manifest["history"]:
                if "v1Compatibility" in entry:
                    entry = json.loads(entry["v1Compatibility"])
                    if "config" in entry:
                        if key in entry["config"]:
                            cmd = entry["config"][key]

    # Standard is to include commands like ['/bin/sh']
    if isinstance(cmd, list):
        if delim is not None:
            cmd = delim.join(cmd)
    bot.verbose("Found Docker config (%s) %s" % (key, cmd))
    return cmd


def get_environment_tar(self):
    """return the environment.tar generated with the Singularity software.
    We first try the Linux Filesystem expected location in /usr/libexec
    If not found, we detect the system archicture

    dirname $(singularity selftest 2>&1 | grep 'lib' | awk '{print $4}' | sed -e 's@\(.*/singularity\).*@\1@')
    """
    from sregistry.utils import which, run_command

    # First attempt - look at File System Hierarchy Standard (FHS)
    res = which("singularity")["message"]
    libexec = res.replace("/bin/singularity", "")
    envtar = "%s/libexec/singularity/bootstrap-scripts/environment.tar" % libexec

    if os.path.exists(envtar):
        return envtar

    # Second attempt, debian distribution will identify folder
    try:
        res = which("dpkg-architecture")["message"]
        if res is not None:
            cmd = ["dpkg-architecture", "-qDEB_HOST_MULTIARCH"]
            triplet = run_command(cmd)["message"].strip("\n")
            envtar = (
                "/usr/lib/%s/singularity/bootstrap-scripts/environment.tar" % triplet
            )
            if os.path.exists(envtar):
                return envtar
    except:
        pass

    # Final, return environment.tar provided in package
    return "%s/environment.tar" % os.path.abspath(os.path.dirname(__file__))


def create_metadata_tar(self, destination=None, metadata_folder=".singularity.d"):
    """create a metadata tar (runscript and environment) to add to the
    downloaded image. This function uses all functions in this section
    to obtain key--> values from the manifest config, and write
    to a .tar.gz

    Parameters
    ==========
    metadata_folder: the metadata folder in the singularity image.
                     default is .singularity.d
    """
    tar_file = None

    # We will add these files to it
    files = []

    # Extract and add environment
    environ = self._extract_env()
    if environ not in [None, ""]:
        bot.verbose3("Adding Docker environment to metadata tar")
        template = get_template("tarinfo")
        template["name"] = "./%s/env/10-docker.sh" % (metadata_folder)
        template["content"] = environ
        files.append(template)

    # Extract and add labels
    labels = self._extract_labels()
    if labels is not None:
        labels = print_json(labels)
        bot.verbose3("Adding Docker labels to metadata tar")
        template = get_template("tarinfo")
        template["name"] = "./%s/labels.json" % metadata_folder
        template["content"] = labels
        files.append(template)

    # Runscript
    runscript = self._extract_runscript()
    if runscript is not None:
        bot.verbose3("Adding Docker runscript to metadata tar")
        template = get_template("tarinfo")
        template["name"] = "./%s/runscript" % metadata_folder
        template["content"] = runscript
        files.append(template)

    if len(files) > 0:
        dest = self._get_download_cache(destination, subfolder="metadata")
        tar_file = create_tar(files, dest)
    else:
        bot.warning("No metadata will be included.")
    return tar_file


def extract_env(self):
    """extract the environment from the manifest, or return None.
    Used by functions env_extract_image, and env_extract_tar
    """
    environ = self._get_config("Env")
    if environ is not None:
        if not isinstance(environ, list):
            environ = [environ]

        lines = []
        for line in environ:
            line = re.findall("(?P<var_name>.+?)=(?P<var_value>.+)", line)
            line = ['export %s="%s"' % (x[0], x[1]) for x in line]
            lines = lines + line

        environ = "\n".join(lines)
        bot.verbose3("Found Docker container environment!")
    return environ


def extract_runscript(self):
    """extract the runscript (EntryPoint) as first priority, unless the
    user has specified to use the CMD. If Entrypoint is not defined,
    we default to None:

    1. IF SREGISTRY_DOCKERHUB_CMD is set, use Cmd
    2. If not set, or Cmd is None/blank, try Entrypoint
    3. If Entrypoint is not set, use default /bin/bash

    """
    use_cmd = self._get_setting("SREGISTRY_DOCKER_CMD")

    # Does the user want to use the CMD instead of ENTRYPOINT?
    commands = ["Entrypoint", "Cmd"]
    if use_cmd is not None:
        commands.reverse()

    # Parse through commands until we hit one
    for command in commands:
        cmd = self._get_config(command)
        if cmd is not None:
            break

    # Only continue if command still isn't None
    if cmd is not None:
        bot.verbose3("Adding Docker %s as Singularity runscript..." % command.upper())

        # If the command is a list, join. (eg ['/usr/bin/python','hello.py']
        bot.debug(cmd)
        if not isinstance(cmd, list):
            cmd = [cmd]

        cmd = " ".join(['"%s"' % x for x in cmd])

        cmd = 'exec %s "$@"' % cmd
        cmd = "#!/bin/sh\n\n%s\n" % cmd
        return cmd

    bot.debug("CMD and ENTRYPOINT not found, skipping runscript.")
    return cmd


def extract_labels(self):
    """extract_labels will write a file of key value pairs including
       maintainer, and labels.

    Parameters
    ==========
    manifest: the manifest to use

    """
    labels = self._get_config("Labels")
    if labels in [[], "", None]:
        labels = None

    return labels



"""

push.py: push functions for sregistry client

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot, ProgressBar
from sregistry.utils import parse_image_name, remove_uri
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor

import requests
import sys
import os


def push(self, path, name, tag=None):
    """push an image to Singularity Registry"""

    path = os.path.abspath(path)
    bot.debug("PUSH %s" % path)

    if not os.path.exists(path):
        bot.exit("%s does not exist." % path)

    # Extract the metadata
    names = parse_image_name(remove_uri(name), tag=tag)

    # COLLECTION ###################################################################

    # If the registry is provided in the uri, use it
    if names["registry"] is None:
        names["registry"] = self.base

    # If the registry is still None, no go
    if names["registry"] is None:
        bot.exit("You must define a base in secrets, image uri, or environment")

    # If the base doesn't start with http or https, add it
    names = self._add_https(names)

    # Prepare push request, this will return a collection ID if permission
    url = "%s/push/" % names["registry"]
    auth_url = "%s/upload/chunked_upload" % names["registry"]
    SREGISTRY_EVENT = self.authorize(request_type="push", names=names)

    # Data fields for collection
    fields = {
        "collection": names["collection"],
        "name": names["image"],
        "tag": names["tag"],
    }

    headers = {"Authorization": SREGISTRY_EVENT}

    r = requests.post(auth_url, json=fields, headers=headers)

    # Always tell the user what's going on!
    message = self._read_response(r)
    print("\n[1. Collection return status {0} {1}]".format(r.status_code, message))

    # Get the collection id, if created, and continue with upload
    if r.status_code != 200:
        sys.exit(1)

    # UPLOAD #######################################################################

    url = "%s/upload" % names["registry"].replace("/api", "")
    bot.debug("Setting upload URL to {0}".format(url))

    cid = r.json()["cid"]
    upload_to = os.path.basename(names["storage"])

    SREGISTRY_EVENT = self.authorize(request_type="upload", names=names)

    encoder = MultipartEncoder(
        fields={
            "SREGISTRY_EVENT": SREGISTRY_EVENT,
            "name": names["image"],
            "collection": str(cid),
            "tag": names["tag"],
            "file1": (upload_to, open(path, "rb"), "text/plain"),
        }
    )

    progress_callback = create_callback(encoder, self.quiet)
    monitor = MultipartEncoderMonitor(encoder, progress_callback)
    headers = {"Content-Type": monitor.content_type, "Authorization": SREGISTRY_EVENT}

    try:
        r = requests.post(url, data=monitor, headers=headers)
        r.raise_for_status()
        message = r.json()["message"]
        print("\n[Return status {0} {1}]".format(r.status_code, message))
    except requests.HTTPError as e:
        print("\nUpload failed: {0}.".format(e))
    except KeyboardInterrupt:
        print("\nUpload cancelled.")


def create_callback(encoder, quiet=False):
    encoder_len = encoder.len / (1024 * 1024.0)

    bar = ProgressBar(expected_size=encoder_len, filled_char="=", hide=quiet)

    def callback(monitor):
        bar.show(monitor.bytes_read / (1024 * 1024.0))

    return callback



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
from sregistry.utils import parse_image_name, remove_uri

# see the registry push for example of how to do this
import dropbox
import sys
import os


def push(self, path, name, tag=None):
    """push an image to your Dropbox

    Parameters
    ==========
    path: should correspond to an absolute image path (or derive it)
    name: should be the complete uri that the user has requested to push.
    tag: should correspond with an image tag. This is provided to mirror Docker

    if the image is less than 150MB, the standard file_upload is used. If
    larger, the image is uploaded in chunks with a progress bar.

    """
    path = os.path.abspath(path)
    bot.debug("PUSH %s" % path)

    if not os.path.exists(path):
        bot.exit("%s does not exist." % path)

    # here is an exampole of getting metadata for a container
    names = parse_image_name(remove_uri(name), tag=tag)

    # Get the size of the file
    file_size = os.path.getsize(path)
    chunk_size = 4 * 1024 * 1024
    storage_path = "/%s" % names["storage"]

    # This is MB
    # image_size = os.path.getsize(path) >> 20

    # prepare the progress bar
    progress = 0
    bot.show_progress(progress, file_size, length=35)

    # If image is smaller than 150MB, use standard upload
    with open(path, "rb") as F:
        if file_size <= chunk_size:
            self.dbx.files_upload(F.read(), storage_path)

        # otherwise upload in chunks
        else:
            start = self.dbx.files_upload_session_start(F.read(chunk_size))
            cursor = dropbox.files.UploadSessionCursor(
                session_id=start.session_id, offset=F.tell()
            )
            commit = dropbox.files.CommitInfo(path=storage_path)

            while F.tell() < file_size:
                progress += chunk_size

                # Finishing up the file, less than chunk_size to go
                if file_size - F.tell() <= chunk_size:
                    self.dbx.files_upload_session_finish(
                        F.read(chunk_size), cursor, commit
                    )

                # Finishing up the file, less than chunk_size to go
                else:
                    self.dbx.files_upload_session_append(
                        F.read(chunk_size), cursor.session_id, cursor.offset
                    )
                    cursor.offset = F.tell()

                # Update the progress bar
                bot.show_progress(
                    iteration=progress,
                    total=file_size,
                    length=35,
                    carriage_return=False,
                )

    # Finish up
    bot.show_progress(
        iteration=file_size, total=file_size, length=35, carriage_return=True
    )

    # Newline to finish download
    sys.stdout.write("\n")



"""

search and query functions for client

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
from sregistry.utils import parse_image_name, remove_uri

import sys


def search(self, query=None, args=None):
    """query a Singularity registry for a list of images.
     If query is None, collections are listed.

    EXAMPLE QUERIES:

    [empty]             list all collections in registry
    vsoch               do a general search for the expression "vsoch"
    vsoch/              list all containers in collection vsoch
    /dinosaur           list containers across collections called "dinosaur"
    vsoch/dinosaur      list details of container vsoch/dinosaur
                          tag "latest" is used by default, and then the most recent
    vsoch/dinosaur:tag  list details for specific container

    """

    if query is not None:
        # List all containers in collection query/
        if query.endswith("/"):  # collection search
            return self._collection_search(query)

        # List containers across collections called /query
        elif query.startswith("/"):
            return self._container_search(query, across_collections=True)

        # List details of a specific collection container
        elif "/" in query or ":" in query:
            return self._container_search(query)

        # Search collections across all fields
        return self._collection_search(query=query)

    # Search collections across all fields
    return self._search_all()


##################################################################
# Search Helpers
##################################################################


def search_all(self):
    """a "show all" search that doesn't require a query"""

    url = "%s/collections/" % self.base

    results = self._paginate_get(url)

    if len(results) == 0:
        bot.info("No container collections found.")
        sys.exit(1)

    bot.info("Collections")

    rows = []
    for result in results:
        if "containers" in result:
            for c in result["containers"]:
                rows.append([c["uri"], c["detail"]])

    bot.table(rows)
    return rows


def collection_search(self, query):
    """collection search will list all containers for a specific
    collection. We assume query is the name of a collection"""

    query = query.lower().strip("/")
    url = "%s/collection/%s" % (self.base, query)

    result = self._get(url)
    if len(result) == 0:
        bot.info("No collections found.")
        sys.exit(1)

    bot.custom(prefix="COLLECTION", message=query)

    rows = []
    for container in result["containers"]:
        rows.append([container["uri"], container["detail"]])

    bot.table(rows)
    return rows


def label_search(self, key=None, value=None):
    """search across labels"""

    if key is not None:
        key = key.lower()

    if value is not None:
        value = value.lower()

    show_details = True
    if key is None and value is None:
        url = "%s/labels/search" % (self.base)
        show_details = False

    elif key is not None and value is not None:
        url = "%s/labels/search/%s/key/%s/value" % (self.base, key, value)

    elif key is None:
        url = "%s/labels/search/%s/value" % (self.base, value)

    else:
        url = "%s/labels/search/%s/key" % (self.base, key)

    result = self._get(url)
    if len(result) == 0:
        bot.info("No labels found.")
        sys.exit(0)

    bot.info("Labels\n")

    rows = []
    for l in result:
        if show_details is True:
            entry = [
                "%s:%s" % (l["key"], l["value"]),
                "\n%s\n\n" % "\n".join(l["containers"]),
            ]
        else:
            entry = ["N=%s" % len(l["containers"]), "%s:%s" % (l["key"], l["value"])]
        rows.append(entry)
    bot.table(rows)
    return rows


def container_search(self, query, across_collections=False):
    """search for a specific container. If across collections is False,
    the query is parsed as a full container name and a specific container
    is returned. If across_collections is True, the container is searched
    for across collections. If across collections is True, details are
    not shown"""

    query = query.lower().strip("/")

    q = parse_image_name(remove_uri(query), defaults=False)

    if q["tag"] is not None:
        if across_collections is True:
            url = "%s/container/search/name/%s/tag/%s" % (
                self.base,
                q["image"],
                q["tag"],
            )
        else:
            url = "%s/container/search/collection/%s/name/%s/tag/%s" % (
                self.base,
                q["collection"],
                q["image"],
                q["tag"],
            )

    elif q["tag"] is None:
        if across_collections is True:
            url = "%s/container/search/name/%s" % (self.base, q["image"])
        else:
            url = "%s/container/search/collection/%s/name/%s" % (
                self.base,
                q["collection"],
                q["image"],
            )

    result = self._get(url)
    if "containers" in result:
        result = result["containers"]

    if len(result) == 0:
        bot.info("No containers found.")
        sys.exit(1)

    bot.info("Containers %s" % query)

    rows = []
    for c in result:
        rows.append(["%s/%s" % (c["collection"], c["name"]), c["tag"]])

    bot.table(rows)
    return rows



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
from sregistry.utils import remove_uri


def main(args, parser, extra):
    from sregistry.main import get_client

    for query in args.query:
        original = query
        query = remove_uri(query)

        if query in ["", "*"]:
            query = None

        try:
            cli = get_client(original, quiet=args.quiet)
            cli.announce(args.command)
            cli.search(query, args=args)
        except NotImplementedError:
            msg = "search is not implemented for %s. Why don't you add it?"
            bot.exit(msg % cli.client_name)



"""

Copyright (C) 2018-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
from sregistry.utils import (
    get_file_hash,
    get_recipe_tag,
    get_tmpdir,
    parse_image_name,
    remove_uri,
)

from sregistry.main.google_build.utils import get_build_template

try:
    from urllib.parse import unquote
except:
    from urllib2 import unquote

from glob import glob
import time
import tarfile
import shutil
import json
import os
import re

################################################################################
# Manual Build Logic
# 1. Intended for run on the command line, simply use the client to run sregistry
#    build, provide a recipe and a name, and it will update the terminal with
#    status until finished using the build function here.
#     - If you provide a GitHub uri in the recipe, instead we run build_repo
#       and get the recipe file from there.
#     - If headless is False, we assume running from the sregistry client and
#       wait for the build to finish, returning the URL to the user.
#     - If headless is True, we assume the function is being run outside of
#       the client, and build_status and build_finish can be used for updates.
#     - if a webhook is provided, headless must be True, and a response is
#       sent here with curl (under development)


def build(
    self,
    name,
    recipe="Singularity",
    context=None,
    preview=False,
    headless=False,
    working_dir=None,
    webhook=None,
    timeout=10800,
    extra_data=None,
):
    """trigger a build on Google Cloud (builder then storage) given a name
    recipe, and Github URI where the recipe can be found. This means
    creating and uploading a build package to use for the build.

    Parameters
    ==========
    recipe: the local recipe to build.
    name: should be the complete uri that the user has requested to push.
    context: the dependency files needed for the build. If not defined, only
             the recipe is uploaded.
    preview: if True, preview but don't run the build
    working_dir: The working directory for the build. Defaults to pwd.
    timeout: the number of seconds for the build to timeout. The default
             is 3 hours, and the maximum is 24 hours. If unset (None)
             it will be 10 minutes.
    webhook: if not None, add a curl POST to finish the build.
    headless: If true, don't track the build, but submit and provide
              an endpoint to send a response to.
    extra_data: a dictionary of extra_data to send back to the webhook (they
                are passed in the environment)

    Environment
    ===========
    SREGISTRY_GOOGLE_BUILD_SINGULARITY_VERSION: the version of Singularity
        to use, defaults to v3.2.1-slim
    SREGISTRY_GOOGLE_BUILD_CLEANUP: after build, delete intermediate
        dependencies in cloudbuild bucket.

    """
    bot.debug("BUILD %s" % recipe)

    build_package = [recipe]

    if context:
        # If the user gives a ., include recursive $PWD
        if "." in context:
            context = glob(os.getcwd() + "/**/*", recursive=True)
        build_package = build_package + context

    # We need to get and save relative paths for the config.
    package = create_build_package(build_package, working_dir)

    # Does the package already exist? If the user cached, it might
    destination = "source/%s" % os.path.basename(package)
    blob = self._build_bucket.blob(destination)

    # if it doesn't exist, upload it
    if not blob.exists() and preview is False:
        bot.log("Uploading build package!")
        self._upload(source=package, bucket=self._build_bucket, destination=destination)
    else:
        bot.log("Build package found in %s." % self._build_bucket.name)

    # This returns a data structure with collection, container, based on uri
    names = parse_image_name(remove_uri(name))

    prefix = "%s/" % names["collection"]

    # The name should include the complete uri so it's searchable
    name = os.path.basename(names["uri"])

    # Update the recipe name to use a relative path
    recipe = get_relative_path(recipe, working_dir)

    # Load the build configuration (defaults to local)
    config = self._load_build_config(name=name, prefix=prefix, recipe=recipe)
    # Add a webhook, if defined
    if webhook and headless:
        config = add_webhook(config=config, webhook=webhook, extra_data=extra_data)

    # The source should point to the bucket with the .tar.gz, latest generation
    config["source"]["storageSource"]["bucket"] = self._build_bucket.name
    config["source"]["storageSource"]["object"] = destination

    # If the user wants a timeout
    if timeout is not None:
        config["timeout"] = "%ss" % timeout

    # If not a preview, run the build and return the response
    if not preview:
        if not headless:
            config = self._run_build(config)
        else:
            config = self._submit_build(config)

        # If the user wants to cache cloudbuild files, this will be set
        env = "SREGISTRY_GOOGLE_BUILD_CACHE"
        if not self._get_and_update_setting(env, self.envars.get(env)):
            if headless is False:
                blob.delete()

    # Clean up either way, return config or response
    shutil.rmtree(os.path.dirname(package))
    return config


################################################################################
# GitHub Build Logic
# 1. Intended for run from a server, we submit a job that uses a Github repo#
#    as a parameter, and if there is a response url provided, the finished
#    build will send a response to that server with curl (recommended).
#    If a private repository is provided, it must be prefixed with a token


def build_repo(
    self,
    repo,
    recipe,
    commit=None,
    branch=None,
    headless=False,
    preview=False,
    webhook=None,
    timeout=10800,
    token=None,
    extra_data=None,
):
    """trigger a build on Google Cloud (builder then storage) given a
    Github repo where a recipe can be found. We assume that github.com (or
    some other Git repo) is provided in the name (repo).

    Parameters
    ==========
    repo: the full repository address
    recipe: the local recipe to build.
    headless: if True, return the first response (and don't wait to finish)
    commit: if not None, check out a commit after clone.
    branch: if defined, checkout a branch.
    token: if an authentication token is provided, add to GitHub clone url.
    timeout: the number of seconds for the build to timeout. The default
             is 3 hours, and the maximum is 24 hours. If unset (None)
             it will be 10 minutes.
    webhook: if not None, add a curl POST to finish the build.
    extra_data: a dictionary of extra_data to send back to the webhook (they
                are passed in the environment)
    """
    bot.debug("BUILD %s" % recipe)

    # This returns a data structure with collection, container, based on uri
    names = parse_image_name(remove_uri(repo))

    # Strip git from the end
    if repo.endswith("git"):
        repo = repo.rstrip(".git")

    # In case they added a tag, strip
    if not repo.startswith("http") and not repo.startswith("git@"):
        repo = "https://%s" % repo

    bot.debug("REPO %s" % repo)

    # First preference to command line, then recipe tag
    tag = get_recipe_tag(recipe) or names.get("tag")

    # Update the tag, if recipe provides one
    names = parse_image_name(remove_uri(repo), tag=tag)

    # <collection>/<image>/<tag>/<commit|branch>/<container>.sif
    prefix = "%s/" % names["collection"]

    # The name should include the complete uri so it's searchable
    name = os.path.basename(names["uri"])

    if commit or branch:
        name = "%s@%s" % (name, (commit or branch))

    # Load the build configuration
    config = self._load_build_config(
        template="singularity-cloudbuild-git.json",
        prefix=prefix,
        name=name,
        recipe=recipe,
    )

    # If we need to checkout a particular commit
    if commit or branch:
        commit = commit or branch
        config["steps"].insert(
            0, {"args": ["checkout", commit], "name": "gcr.io/cloud-builders/git"}
        )

    # If the repo is private, add the token
    clone_url = repo
    if token is not None:
        clone_url = "https://%s@%s" % (token, remove_uri(repo))

    # Add the GitHub repo to the recipe, clone to $PWD (.)
    config["steps"].insert(
        0, {"args": ["clone", clone_url, "."], "name": "gcr.io/cloud-builders/git"}
    )

    # If the user wants a timeout
    if timeout is not None:
        config["timeout"] = "%ss" % timeout

    # Add the webhook step, if applicable.
    if webhook and headless:
        config = add_webhook(config=config, webhook=webhook, extra_data=extra_data)

    # If not a preview, run the build and return the response
    if not preview:
        if not headless:
            return self._run_build(config)
        return self._submit_build(config)

    # If preview is True, we return the config
    return config


def create_build_package(package_files, working_dir=None):
    """given a list of files, copy them to a temporary folder,
    compress into a .tar.gz, and rename based on the file hash.
    Return the full path to the .tar.gz in the temporary folder.

    Parameters
    ==========
    package_files: a list of files to include in the tar.gz
    working_dir: If set, the path derived for the recipe and
                 files is relative to this.

    """
    # Ensure package files all exist
    for package_file in package_files:
        if not os.path.exists(package_file):
            bot.exit("Cannot find %s." % package_file)

    bot.log("Generating build package for %s files..." % len(package_files))
    build_dir = get_tmpdir(prefix="sregistry-build")
    build_tar = "%s/build.tar.gz" % build_dir
    tar = tarfile.open(build_tar, "w:gz")

    # Create the tar.gz, making sure relative to working_dir
    for package_file in package_files:
        # Get a relative path
        relative_path = get_relative_path(package_file, working_dir)
        tar.add(package_file, arcname=relative_path)
    tar.close()

    # Get hash (sha256), and rename file
    sha256 = get_file_hash(build_tar)
    hash_tar = "%s/%s.tar.gz" % (build_dir, sha256)
    shutil.move(build_tar, hash_tar)
    return hash_tar


def get_relative_path(filename, working_dir=None):
    """given a filename and a working directory, return
    a relative path for the builder to use. This means
    removing the working directory from the filename

    Parameters
    ==========
    filename: the name of the file to get a path for
    working_dir: if not None, remove from paths.
    """
    relative_path = filename

    # If a working directory is provided, it must end with /
    if working_dir is not None:
        if not working_dir.endswith(os.sep):
            working_dir = "%s%s" % (working_dir, os.sep)

    if working_dir is not None:
        relative_path = filename.replace(working_dir, "")
    return relative_path


def add_webhook(config, webhook, extra_data=None):
    """add a webhook to a config. We assume that the sha256 is calculated in
    the present working directory. Optionally, the user can provide
    extra_data to post back with the build_id.
    """
    data = {"id": "$BUILD_ID"}
    substitutions = {}

    if extra_data is not None:
        # Keep a list of envars to add
        for key, val in extra_data.items():
            sub = "_SREGISTRY_%s" % key.upper()
            data[key] = "${%s}" % sub
            substitutions[sub] = val

    config["steps"].append(
        {
            "name": "gcr.io/cloud-builders/curl",
            "args": ["-d", json.dumps(data), "-X", "POST", webhook],
        }
    )

    config["substitutions"] = substitutions
    return config


def load_build_config(
    self,
    name,
    recipe,
    template="singularity-cloudbuild-local.json",
    version="v3.2.1-slim",
    prefix="",
):
    """load a google compute config, meaning that we start with a template,
    and mimic the following example cloudbuild.yaml:

     steps:
     - name: "singularityware/singularity:${_SINGULARITY_VERSION}"
       args: ['build', 'julia-centos-another.sif', 'julia.def']
     artifacts:
       objects:
         location: 'gs://sregistry-gcloud-build-vanessa'
         paths: ['julia-centos-another.sif']


     Parameters
     ==========
     recipe: the local recipe file for the builder.
     name: the name of the container, based on the uri
     template: the template to use, will be populated based on name
     prefix: a prefix for the storage path, defaults to empty string.
    """
    version_envar = "SREGISTRY_GOOGLE_BUILD_SINGULARITY_VERSION"
    version = self._get_and_update_setting(version_envar, version)

    # Get the build template based on its name
    config = get_build_template(template)

    # Name is in format 'dinosaur/container:latest'
    storage_name = "%s%s.sif" % (prefix, name)
    container_name = os.path.basename(storage_name)

    # We need to create the equivalent directory for the image
    folder_name = os.path.dirname(storage_name)

    # The bucket location should have the uri
    bucket_location = "gs://%s/%s/" % (self._bucket_name, folder_name)

    # Last step: tag the blog with the sha256 sum
    bucket_sif = "%s%s" % (bucket_location, container_name)
    metadata_command = (
        "gsutil setmeta -h x-goog-meta-sha256sum:$(cat SHA256SUM) %s" % bucket_sif
    )
    config["steps"].insert(
        0,
        {
            "args": ["-c", metadata_command],
            "entrypoint": "/bin/bash",
            "name": "gcr.io/cloud-builders/gsutil",
        },
    )

    # Second to last step: manually upload the blob
    config["steps"].insert(
        0,
        {
            "args": ["cp", container_name, bucket_sif],
            "name": "gcr.io/cloud-builders/gsutil",
        },
    )

    # Calculate a hash (happens after build)
    sha256_command = "sha256sum %s | cut -c 1-64 > SHA256SUM" % container_name
    config["steps"].insert(
        0, {"args": ["-c", sha256_command], "entrypoint": "/bin/bash", "name": "ubuntu"}
    )

    # Insert the build step (second step)
    config["steps"].insert(
        0,
        {
            "args": ["build", container_name, recipe],
            "name": "singularityware/singularity:%s" % version,
        },
    )

    return config


def run_build(self, config):
    """run a build, meaning creating a build. Retry if there is failure"""

    response = self._submit_build(config)
    status = response["metadata"]["build"]["status"]
    build_id = response["metadata"]["build"]["id"]

    start = time.time()
    while status not in ["COMPLETE", "FAILURE", "SUCCESS", "TIMEOUT"]:
        time.sleep(15)
        response = self._build_status(build_id)
        status = response["status"]

    end = time.time()
    bot.log("Total build time: %s seconds" % (round(end - start, 2)))
    return self._finish_build(build_id, response=response, config=config)


def submit_build(self, config):
    """run a build, meaning creating a build. Retry if there is failure"""

    project = self._get_project()

    #          prefix,    message, color
    bot.custom("PROJECT", project, "CYAN")
    for i, step in enumerate(config["steps"]):
        bot.custom("BUILD %s" % i, step["name"], "CYAN")

    response = (
        self._build_service.projects()
        .builds()
        .create(body=config, projectId=project)
        .execute()
    )

    build_id = response["metadata"]["build"]["id"]
    status = response["metadata"]["build"]["status"]
    bot.log("build %s: %s" % (build_id, status))

    return response


def build_status(self, build_id):
    """get a build status based on a build id. We return the entire response
    object for the client to parse.
    """
    project = self._get_project()
    response = (
        self._build_service.projects()
        .builds()
        .get(id=build_id, projectId=project)
        .execute()
    )

    build_id = response["id"]
    status = response["status"]
    bot.log("build %s: %s" % (build_id, status))

    return response


def finish_build(self, build_id, config=None, response=None):
    """finish a build, meaning if the build was successful, we check the
    user preference to make it private. If it's set, we leave it
    private. Otherwise, we make it public (default).

    Parameters
    ==========
    build_id: the build id returned from submission to track the build.
    config: optionally provide a config to get metadata from for the blob
    response: if we've already gotten a response status, include here
    """
    # Get the build status, we will only complete on SUCCESS
    if not response:
        response = self._build_status(build_id)

    # If successful, update blob metadata and visibility
    if response["status"] == "SUCCESS":
        # Does the user want to keep the container private?
        env = "SREGISTRY_GOOGLE_STORAGE_PRIVATE"
        blob_location = get_blob_location(response, self._bucket_name)
        blob = self._bucket.get_blob(blob_location)

        # Make Public, if desired
        if not self._get_and_update_setting(env, self.envars.get(env)):
            blob.make_public()
            response["public_url"] = unquote(blob.public_url)

        # Add the metadata directly to the object, return updaated response
        response = update_blob_metadata(
            blob=blob, response=response, config=config, bucket=self._bucket
        )

    return response


def get_blob_location(response, bucket):
    """return a relative path for a blob based on finding the build step,
    and the container built from it.

    Parameters
    ==========
    response: the response from client._build_status(build_id)
    bucket: the name of the build bucket.

    """
    # Find the build step, it uses singularityware as a builder
    build_step = [
        x
        for x in response["steps"]
        if x["name"] == "gcr.io/cloud-builders/gsutil" and x["args"][0] == "cp"
    ]

    # This is fragile, but we have to get the container name
    if len(build_step) > 0:
        location = build_step[0]["args"][-1]
        return re.sub("(gs://%s|%s)" % (bucket, bucket), "", location).strip("/")
    bot.exit("Cannot find build step with image name.")


def update_blob_metadata(blob, response, bucket, config=None):
    """a specific function to take a blob, along with a SUCCESS response
    from Google build, the original config, and update the blob
    metadata with the artifact file name, dependencies, and image hash.
    """
    metadata = blob.metadata or {}

    metadata.update(
        {
            "crc32c": blob.crc32c,
            "media_link": blob.media_link,
            "self_link": blob.self_link,
            "md5sum": blob.md5_hash,
            "size": blob.size,
            "type": "container",
        }
    )  # identify that the blob is container

    # If a configuration is provided
    if config:
        metadata["buildCommand"] = " ".join(config["steps"][0]["args"])
        metadata["builder"] = config["steps"][0]["name"]

        # If a source was used, add it.
        if "source" in config:
            metadata["storageSourceBucket"] = config["source"]["storageSource"][
                "bucket"
            ]
            metadata["storageSourceObject"] = config["source"]["storageSource"][
                "object"
            ]

    # Update the response, keeping data structure flat (and we don't overwrite)
    response.update(metadata)
    blob.metadata = metadata
    blob._properties["metadata"] = metadata
    blob.patch()
    return response



"""

Copyright (C) 2018-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
from sregistry.utils import parse_image_name, remove_uri

import os
import botocore


def push(self, path, name, tag=None):
    """push an image to an S3 endpoint"""

    path = os.path.abspath(path)
    bot.debug("PUSH %s" % path)

    if not os.path.exists(path):
        bot.exit("%s does not exist." % path)

    # Extract the metadata
    names = parse_image_name(remove_uri(name), tag=tag)
    image_size = os.path.getsize(path) >> 20

    # Create extra metadata, this is how we identify the image later
    # *important* bug in boto3 will return these capitalized
    # see https://github.com/boto/boto3/issues/1709
    metadata = {"sizemb": "%s" % image_size, "client": "sregistry", "type": "container"}

    ExtraArgs = {"Metadata": metadata}

    acl = self._get_and_update_setting("SREGISTRY_S3_OBJECT_ACL")

    if acl is not None:
        ExtraArgs["ACL"] = acl

    try:
        self.bucket.upload_file(path, names["storage"], ExtraArgs)
    except botocore.exceptions.ClientError as e:
        bot.exit(
            "Could not upload {} to bucket. Ensure you have sufficient permissions to put objects in the bucket (s3:PutObject), as well as modify the object ACL if SREGISTRY_S3_OBJECT_ACL is set (s3:PutObjectAcl): {}".format(
                path, str(e)
            )
        )



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

import os


def get_or_create_folder(self, folder):
    """create a folder at the drive root. If the folder already exists,
    it is simply returned.

        folder = self._get_or_create_folder(self._base)
        $ folder
          {'id': '1pXR5S8wufELh9Q-jDkhCoYu-BL1NqN9y'}

    """
    q = "mimeType='application/vnd.google-apps.folder' and name='%s'" % folder
    response = (
        self._service.files().list(q=q, spaces="drive").execute().get("files", [])
    )

    # If no folder is found, create it!
    if len(response) == 0:
        folder = self._create_folder(folder)
    else:
        folder = response[0]
    return folder


def create_folder(self, folder):
    """create a folder with a particular name. Be careful, if the folder
    already exists you can still create one (a different one) with
    the equivalent name!
    """
    folder_metadata = {
        "name": os.path.basename(folder),
        "mimeType": "application/vnd.google-apps.folder",
    }
    created = self._service.files().create(body=folder_metadata, fields="id").execute()
    return created



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
import json
import sys
import re


def main(args, parser, extra):
    from sregistry.main import get_client

    if args.name is None:
        msg = "You must add the --name of a container uri to build."
        bot.exit(msg)

    cli = get_client(image=args.name, quiet=args.quiet)
    cli.announce(args.command)

    # If the client doesn't have the command, exit
    if not hasattr(cli, "build"):
        msg = "build is not implemented for %s. Why don't you add it?"
        bot.exit(msg % cli.client_name)

    # Singularity Registry Server uses build with a recipe
    if cli.client_name == "google-build":
        response = run_google_build(cli, args)

    elif cli.client_name == "google-storage":
        response = run_compute_build(cli, args)

    # Currently allows for google_build
    else:
        bot.warning("No URI specified, assuming Singularity Registry with Builder")
        response = run_registry_build(cli, args, extra)

    # If the client wants to preview, the config is returned
    if args.preview:
        print(json.dumps(response, indent=4, sort_keys=True))


def run_google_build(cli, args):
    """a helper function to control build for the Google Build Client.
    the user can request a local recipe build, or a build from
    a GitHub repo. In both cases, if no recipe is provided, we default
    to Singularity at the root of the PWD or GitHub repository.
    """
    if args.name is None:
        bot.exit("Please provide a container identifier with --name")

    # Default to recipe "Singularity" unless other is provided
    recipe = "Singularity"
    if args.commands:
        recipe = args.commands.pop(0)

    # If Github.com is provided in the name, we are doing a GitHub build
    if re.search("github.com|gitlab.com", args.name):
        name = args.name.replace("google-build://", "")
        response = cli.build_repo(repo=name, recipe=recipe, preview=args.preview)

    else:
        response = cli.build(
            name=args.name, recipe=recipe, context=args.commands, preview=args.preview
        )

    # Print output to the console
    if not args.preview:
        print_output(response, args.outfile)
    return response


def run_compute_build(cli, args):
    """a compute based build is the oldest versions of build - here we bring
    up our own instance, and then provide control to it. The helper
    functions below (kill, instances, templates) support this version.
    """
    # Does the user want to save the image?
    command = args.commands.pop(0)

    # Option 1: The user wants to kill an instance
    if command == "kill":
        kill(args)

    # Option 2: Just list running instances
    elif command == "instances":
        instances(args)

    # Option 3: The user wants to list templates
    elif "template" in command:
        templates(args)

    # Option 4: View a specific or latest log
    elif command == "logs":
        list_logs(args)

    # Option 3: The user is providing a Github repo!
    recipe = "Singularity"

    if "github" in command:
        # One argument indicates a recipe
        if len(args.commands) == 1:
            recipe = args.commands.pop(0)

    else:
        # If a command is provided, but not a Github repo
        bot.exit("%s is not a recognized option." % command)

    # Does the user want to specify a name for the collection?
    name = args.name

    # No image is needed, we are creating in the cloud
    return cli.build(repo=command, name=name, recipe=recipe, preview=args.preview)


def run_registry_build(cli, args, extra):
    """a registry build pushes a recipe file to Singularity Registry Server,
    or given that a GitHub Url is provided, we build from there. For more
    regular building, the user is suggested to directly connect the
    repository to Singularity Registry server. This can serve as a one time
    build.
    """
    # The uri can also contain github, which indicates a Github build
    if args.name is None:
        bot.exit("Please provide a container identifier with --name")

    recipe = args.commands.pop(0)
    response = cli.build(name=args.name, recipe=recipe, extra=extra)

    # Print output to the console
    if response is not None:
        print_output(response, args.outfile)
    return response


def print_output(response, output_file=None):
    """print the output to the console for the user. If the user wants the content
    also printed to an output file, do that.

    Parameters
    ==========
    response: the response from the builder, with metadata added
    output_file: if defined, write output also to file

    """
    # If successful built, show container uri
    if "status" in response:
        if response["status"] == "SUCCESS":
            bucket = response["artifacts"]["objects"]["location"]
            obj = response["artifacts"]["objects"]["paths"][0]
            bot.custom("MD5HASH", response["file_hash"], "CYAN")
            bot.custom("SIZE", response["size"], "CYAN")
            bot.custom(response["status"], bucket + obj, "CYAN")
        else:
            bot.custom(response["status"], "see logs for details", "CYAN")

        # Show the logs no matter what
        bot.custom("LOGS", response["logUrl"], "CYAN")

        # Did the user make the container public?
        if "public_url" in response:
            bot.custom("URL", response["public_url"], "CYAN")

    # Does the user also need writing to an output file?
    if output_file is not None:
        with open(output_file, "w") as filey:
            if response["status"] == "SUCCESS":
                filey.writelines("MD5HASH %s\n" % response["file_hash"])
                filey.writelines("SIZE %s\n" % response["size"])
            filey.writelines("%s %s%s\n" % (response["status"], bucket, obj))
            filey.writelines("LOGS %s\n" % response["logUrl"])
            if "public_url" in response:
                filey.writelines("URL %s\n" % response["public_url"])


def kill(args):
    """kill is a helper function to call the "kill" function of the client,
    meaning we bring down an instance.
    """
    from sregistry.main import get_client

    cli = get_client(quiet=args.quiet)
    if len(args.commands) > 0:
        for name in args.commands:
            cli.destroy(name)
    sys.exit(0)


def instances(args):
    """list running instances for a user, including all builders and report
    instance names and statuses.
    """
    from sregistry.main import get_client

    cli = get_client(quiet=args.quiet)
    cli.list_builders()
    sys.exit(0)


def templates(args, template_name=None):
    """list a specific template (if a name is provided) or all templates
    available.

    Parameters
    ==========
    args: the argparse object to look for a template name
    template_name: if not set, show all

    """
    from sregistry.main import get_client

    # We don't need storage/compute connections
    cli = get_client(init=False, quiet=args.quiet)

    if len(args.commands) > 0:
        template_name = args.commands.pop(0)
    cli.list_templates(template_name)
    sys.exit(0)


def list_logs(args, container_name=None):
    """list a specific log for a builder, or the latest log if none provided

    Parameters
    ==========
    args: the argparse object to look for a container name
    container_name: a default container name set to be None (show latest log)

    """
    from sregistry.main import get_client

    cli = get_client(quiet=args.quiet)
    if args.commands:
        container_name = args.commands.pop(0)
    cli.logs(container_name)
    sys.exit(0)



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
import re


def search(self, query=None, **kwargs):
    """query a Singularity registry for a list of images.
     If query is None, collections are listed.

    EXAMPLE QUERIES:

    [empty]             list all collections in singularity hub
    vsoch               do a general search for collection "vsoch"
    vsoch/dinosaur      list details of container vsoch/dinosaur
                          tag "latest" is used by default, and then the most recent
    vsoch/dinosaur:tag  list details for specific container

    """

    if query is not None:
        return self._search_collection(query)

    # Search collections across all fields
    return self.list()


##################################################################
# Search Helpers
##################################################################


def list_all(self, **kwargs):
    """a "show all" search that doesn't require a query"""

    quiet = False
    if "quiet" in kwargs:
        quiet = kwargs["quiet"]

    bot.spinner.start()
    url = "%s/collections/" % self.base
    results = self._paginate_get(url)
    bot.spinner.stop()

    if len(results) == 0:
        bot.exit("No container collections found.", return_code=0)

    rows = []
    for result in results:
        if "containers" in result:
            if result["id"] not in [37, 38, 39]:
                for c in result["containers"]:
                    rows.append([c["detail"], "%s:%s" % (c["name"], c["tag"])])

    if quiet is False:
        bot.info("Collections")
        bot.table(rows)

    return rows


def search_collection(self, query):
    """collection search will list all containers for a specific
    collection. We assume query is the name of a collection"""

    query = query.lower().strip("/")

    # Workaround for now - the Singularity Hub search endpoind needs fixing
    containers = self.list(quiet=True)

    rows = []
    for result in containers:
        if re.search(query, result[1]):
            rows.append(result)

    if len(rows) > 0:
        bot.table(rows)
    else:
        bot.info("No containers found.")

    return rows



"""

Copyright (C) 2018-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
from sregistry.utils import parse_image_name, remove_uri
from swiftclient.exceptions import ClientException
import os
import sys


def pull(self, images, file_name=None, save=True, **kwargs):
    """pull an image from storage using Swift. The image is found based on the
    storage uri

    Parameters
    ==========
    images: refers to the uri given by the user to pull in the format
    <collection>/<namespace>. You should have an API that is able to
    retrieve a container based on parsing this uri.
    file_name: the user's requested name for the file. It can
               optionally be None if the user wants a default.
    save: if True, you should save the container to the database
          using self.add()

    Returns
    =======
    finished: a single container path, or list of paths

    """
    force = False
    if "force" in kwargs:
        force = kwargs["force"]

    if not isinstance(images, list):
        images = [images]

    bot.debug("Execution of PULL for %s images" % len(images))

    # If used internally we want to return a list to the user.
    finished = []
    for image in images:
        names = parse_image_name(remove_uri(image))

        # First try to get the collection
        collection = self._get_collection(names["collection"])
        if collection is None:
            bot.error("Collection %s does not exist." % names["collection"])

            # Show the user collections he/she does have access to
            collections = self.get_collections()
            if collections:
                bot.info("Collections available to you: \n%s" % "\n".join(collections))
            sys.exit(1)

        # Determine if the container exists in storage
        image_name = os.path.basename(names["storage"])

        try:
            obj_tuple = self.conn.get_object(names["collection"], image_name)
        except ClientException:
            bot.exit("%s does not exist." % names["storage"])

        # Give etag as version if version not defined
        if names["version"] is None:
            names["version"] = obj_tuple[0]["etag"]

        # If the user didn't provide a file, make one based on the names
        if file_name is None:
            file_name = self._get_storage_name(names)

        # If the file already exists and force is False
        if os.path.exists(file_name) and force is False:
            bot.exit("Image exists! Remove first, or use --force to overwrite")

        # Write to file
        with open(file_name, "wb") as filey:
            filey.write(obj_tuple[1])

        # If we save to storage, the uri is the dropbox_path
        if save is True:
            names.update(obj_tuple[0])
            container = self.add(
                image_path=file_name, image_uri=names["uri"], metadata=names
            )

            # When the container is created, this is the path to the image
            image_file = container.image

            if os.path.exists(image_file):
                bot.debug("Retrieved image file %s" % image_file)
                bot.custom(prefix="Success!", message=image_file)
                finished.append(image_file)

            else:
                bot.error(
                    "%s does not exist. Try sregistry search to see images."
                    % image_file
                )

    if len(finished) == 1:
        finished = finished[0]
    return finished



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.utils import parse_image_name, remove_uri
from sregistry.logger import bot


def share(self, query, share_to=None):
    """share will use the client to get a shareable link for an image of choice.
    the functions returns a url of choice to send to a recipient.
    """

    names = parse_image_name(remove_uri(query))

    # Dropbox path is the path in storage with a slash
    dropbox_path = "/%s" % names["storage"]

    # First ensure that exists
    if self.exists(dropbox_path) is True:
        # Create new shared link
        try:
            share = self.dbx.sharing_create_shared_link_with_settings(dropbox_path)

        # Already exists!
        except:
            share = self.dbx.sharing_create_shared_link(dropbox_path)

        bot.info(share.url)
    return share.url



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
from sregistry.utils import read_json, get_installdir
import os


def get_build_template(name="singularity-cloudbuild-local.json"):
    """get default build template.

    Parameters
    ==========
    name: singularity-cloudbuild-local.json (default) that will build a
              container interactively, waiting for the build to finish.
          singularity-cloudbuild-git.json build a recipe from a GitHub
              repository.
    """
    base = get_installdir()
    name = "%s/main/templates/build/%s" % (base, name)

    if os.path.exists(name):
        bot.debug("Found template %s" % name)
        return read_json(name)

    bot.warning("Template %s not found." % name)



"""

Copyright (C) 2017-2022 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

__version__ = "0.2.39"
AUTHOR = "Vanessa Sochat"
AUTHOR_EMAIL = "vsochat@stanford.edu"
NAME = "sregistry"
PACKAGE_URL = "http://www.github.com/singularityhub/sregistry-cli"
KEYWORDS = "singularity containers registry hub"
DESCRIPTION = "Command line tool for working with container storage"
LICENSE = "LICENSE"

################################################################################
# Global requirements


INSTALL_REQUIRES = (
    ("spython", {"min_version": "0.0.63"}),
    ("requests", {"min_version": "2.18.4"}),
    ("pygments", {"min_version": "2.1.3"}),
)

TESTS_REQUIRES = (("pytest", {"min_version": "4.6.2"}),)

################################################################################
# Submodule Requirements (no database, just client)


INSTALL_BASIC_REGISTRY = (
    ("requests-toolbelt", {"min_version": "0.8.0"}),
    ("python-dateutil", {"min_verison": "2.5.3"}),
)

INSTALL_BASIC_AWS = (("awscli", {"min_version": "1.16.19"}),)

INSTALL_BASIC_SWIFT = (
    ("python-swiftclient", {"min_version": "3.6.0"}),
    ("python-keystoneclient", {"min_version": "3.5.0"}),
    ("oauth2client", {"min_version": "3.0"}),
)
INSTALL_BASIC_DROPBOX = (("dropbox", {"min_version": "8.5.1"}),)

INSTALL_BASIC_GLOBUS = (
    ("oauth2client", {"min_version": "3.0"}),
    ("globus_sdk[jwt]", {"min_version": "1.5.0"}),
)

INSTALL_BASIC_GOOGLE_STORAGE = (
    ("oauth2client", {"min_version": "3.0"}),
    ("google-cloud-storage", {"min_version": "1.4.0"}),
    ("google-api-python-client", {"min_version": "1.6.4"}),
    ("retrying", {"min_version": "1.3.3"}),
    ("httplib2", {"min_version": "0.18.0"}),
)

INSTALL_BASIC_GOOGLE_BUILD = (
    ("oauth2client", {"min_version": "3.0"}),
    ("google-cloud-storage", {"min_version": "1.4.0"}),
    ("google-api-python-client", {"min_version": "1.6.4"}),
    ("retrying", {"min_version": "1.3.3"}),
    ("httplib2", {"min_version": "0.18.0"}),
)

INSTALL_BASIC_GOOGLE_DRIVE = (
    ("google-api-python-client", {"min_version": "1.6.4"}),
    ("oauth2client", {"min_version": "3.0"}),
    ("httplib2", {"min_version": "0.18.0"}),
)

INSTALL_BASIC_GOOGLE_COMPUTE = (
    ("oauth2client", {"min_version": "3.0"}),
    ("google-api-python-client", {"min_version": "1.6.4"}),
    ("google-cloud-storage", {"min_version": "1.4.0"}),
    ("httplib2", {"min_version": "0.18.0"}),
)

INSTALL_BASIC_S3 = (("boto3", {"min_version": "1.7.83"}),)

INSTALL_BASIC_ALL = (
    INSTALL_REQUIRES
    + INSTALL_BASIC_S3
    + INSTALL_BASIC_AWS
    + INSTALL_BASIC_SWIFT
    + INSTALL_BASIC_DROPBOX
    + INSTALL_BASIC_GLOBUS
    + INSTALL_BASIC_REGISTRY
    + INSTALL_BASIC_GOOGLE_STORAGE
    + INSTALL_BASIC_GOOGLE_BUILD
    + INSTALL_BASIC_GOOGLE_COMPUTE
    + INSTALL_BASIC_GOOGLE_DRIVE
)

################################################################################
# Submodule Requirements (versions that include database)


INSTALL_REQUIRES_REGISTRY = (
    ("requests-toolbelt", {"min_version": "0.8.0"}),
    ("python-dateutil", {"min_verison": "2.5.3"}),
    ("sqlalchemy", {"max_version": "1.4.48"}),
)

INSTALL_REQUIRES_AWS = (
    ("awscli", {"min_version": "1.16.19"}),
    ("sqlalchemy", {"max_version": "1.4.48"}),
)

INSTALL_REQUIRES_SWIFT = (
    ("python-swiftclient", {"min_version": "3.6.0"}),
    ("sqlalchemy", {"max_version": "1.4.48"}),
    ("oauth2client", {"min_version": "3.0"}),
)

INSTALL_REQUIRES_DROPBOX = (
    ("dropbox", {"min_version": "8.5.1"}),
    ("sqlalchemy", {"max_version": "1.4.48"}),
)

INSTALL_REQUIRES_GLOBUS = (
    ("oauth2client", {"min_version": "3.0"}),
    ("globus_sdk[jwt]", {"min_version": "1.5.0"}),
    ("sqlalchemy", {"max_version": "1.4.48"}),
)

INSTALL_REQUIRES_GOOGLE_STORAGE = (
    ("oauth2client", {"min_version": "3.0"}),
    ("google-cloud-storage", {"min_version": "1.4.0"}),
    ("google-api-python-client", {"min_version": "1.6.4"}),
    ("retrying", {"exact_version": "1.3.3"}),
    ("sqlalchemy", {"max_version": "1.4.48"}),
    ("httplib2", {"min_version": "0.18.0"}),
)

INSTALL_REQUIRES_GOOGLE_BUILD = (
    ("oauth2client", {"min_version": "3.0"}),
    ("google-cloud-storage", {"min_version": "1.4.0"}),
    ("google-api-python-client", {"min_version": "1.6.4"}),
    ("retrying", {"exact_version": "1.3.3"}),
    ("sqlalchemy", {"max_version": "1.4.48"}),
    ("httplib2", {"min_version": "0.18.0"}),
)

INSTALL_REQUIRES_GOOGLE_DRIVE = (
    ("oauth2client", {"min_version": "3.0"}),
    ("sqlalchemy", {"max_version": "1.4.48"}),
    ("google-api-python-client", {"min_version": "1.6.4"}),
    ("httplib2", {"min_version": "0.18.0"}),
)

INSTALL_REQUIRES_GOOGLE_COMPUTE = (
    ("oauth2client", {"min_version": "3.0"}),
    ("sqlalchemy", {"max_version": "1.4.48"}),
    ("google-api-python-client", {"min_version": "1.6.4"}),
    ("google-cloud-storage", {"min_version": "1.4.0"}),
    ("httplib2", {"min_version": "0.18.0"}),
)

INSTALL_REQUIRES_S3 = (
    ("sqlalchemy", {"max_version": "1.4.48"}),
    ("boto3", {"min_version": "1.7.83"}),
)

INSTALL_REQUIRES_ALL = (
    INSTALL_REQUIRES
    + INSTALL_REQUIRES_AWS
    + INSTALL_REQUIRES_S3
    + INSTALL_REQUIRES_SWIFT
    + INSTALL_REQUIRES_DROPBOX
    + INSTALL_REQUIRES_REGISTRY
    + INSTALL_REQUIRES_GOOGLE_COMPUTE
    + INSTALL_REQUIRES_GOOGLE_STORAGE
    + INSTALL_REQUIRES_GOOGLE_BUILD
    + INSTALL_REQUIRES_GOOGLE_DRIVE
)



"""

Copyright (C) 2016-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot


def main(args, parser, extra):
    from sregistry.main import get_client

    image = args.image
    name = args.name

    # Customize client based on uri
    cli = get_client(image, quiet=args.quiet)
    cli.announce(args.command)

    # Does the user want to save the image?
    do_save = True
    if args.nocache is True or not hasattr(cli, "storage"):
        do_save = False

    # If the client doesn't have the command, exit
    if not hasattr(cli, "pull"):
        msg = "pull is not implemented for %s. Why don't you add it?"
        bot.exit(msg % cli.client_name)

    cli.pull(images=image, file_name=name, force=args.force, save=do_save)



from sregistry.defaults import SREGISTRY_DATABASE

if SREGISTRY_DATABASE is None:
    from .dummy import add, init_db
else:
    from .models import Collection, Container, init_db, Base
    from .sqlite import (
        add,
        cp,
        get,
        mv,
        rm,
        images,
        inspect,
        get_container,
        get_collection,
        get_or_create_collection,
        rename,
    )



"""

This is a base client that imports functions depending on the API it is 
    using. Currently, it supports singularity hub and registry, with default
    to use Singularity Hub.

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.main.base import ApiConnection

from sregistry.utils import check_install, get_uri
from sregistry.auth import get_credential_cache
from sregistry.defaults import SREGISTRY_DATABASE
from sregistry.logger import bot
import os


def get_client(image=None, quiet=False, **kwargs):
    """
    get the correct client depending on the driver of interest. The
    selected client can be chosen based on the environment variable
    SREGISTRY_CLIENT, and later changed based on the image uri parsed
    If there is no preference, the default is to load the singularity
    hub client.

    Parameters
    ==========
    image: if provided, we derive the correct client based on the uri
    of an image. If not provided, we default to environment, then hub.
    quiet: if True, suppress most output about the client (e.g. speak)

    """
    from sregistry.defaults import SREGISTRY_CLIENT

    # Give the user a warning:
    if not check_install():
        bot.warning("Singularity is not installed, function might be limited.")

    # If an image is provided, use to determine client
    client_name = get_uri(image)
    if client_name is not None:
        SREGISTRY_CLIENT = client_name

    # If no obvious credential provided, we can use SREGISTRY_CLIENT
    if SREGISTRY_CLIENT == "aws":
        from .aws import Client
    elif SREGISTRY_CLIENT == "docker":
        from .docker import Client
    elif SREGISTRY_CLIENT == "dropbox":
        from .dropbox import Client
    elif SREGISTRY_CLIENT == "gitlab":
        from .gitlab import Client
    elif SREGISTRY_CLIENT == "globus":
        from .globus import Client
    elif SREGISTRY_CLIENT == "nvidia":
        from .nvidia import Client
    elif SREGISTRY_CLIENT == "hub":
        from .hub import Client
    elif SREGISTRY_CLIENT == "google-drive":
        from .google_drive import Client
    elif SREGISTRY_CLIENT == "google-compute":
        from .google_storage import Client
    elif SREGISTRY_CLIENT == "google-storage":
        from .google_storage import Client
    elif SREGISTRY_CLIENT == "google-build":
        from .google_build import Client
    elif SREGISTRY_CLIENT == "registry":
        from .registry import Client
    elif SREGISTRY_CLIENT == "s3":
        from .s3 import Client
    elif SREGISTRY_CLIENT == "swift":
        from .swift import Client
    else:
        from .hub import Client

    Client.client_name = SREGISTRY_CLIENT
    Client.quiet = quiet

    # Create credentials cache, if it doesn't exist
    Client._credential_cache = get_credential_cache()

    # Add the database, if wanted
    if SREGISTRY_DATABASE is not None:
        # These are global functions used across modules
        from sregistry.database import (
            init_db,
            add,
            cp,
            get,
            mv,
            rm,
            images,
            inspect,
            rename,
            get_container,
            get_collection,
            get_or_create_collection,
        )

        # Actions
        Client._init_db = init_db
        Client.add = add
        Client.cp = cp
        Client.get = get
        Client.inspect = inspect
        Client.mv = mv
        Client.rename = rename
        Client.rm = rm
        Client.images = images

        # Collections
        Client.get_or_create_collection = get_or_create_collection
        Client.get_container = get_container
        Client.get_collection = get_collection

    # If no database, import dummy functions that return the equivalent
    else:
        from sregistry.database import add, init_db

        Client.add = add
        Client._init_db = init_db

    # Initialize the database
    cli = Client()

    if hasattr(Client, "_init_db"):
        cli._init_db(SREGISTRY_DATABASE)
    return cli



"""

Copyright (C) 2018-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
from sregistry.main import ApiConnection
import re

from .pull import pull
from .query import search, search_all


class Client(ApiConnection):
    def __init__(self, secrets=None, base=None, **kwargs):
        super(Client, self).__init__(**kwargs)
        self._reset_headers()
        self._update_secrets()
        self._update_base()

    def _update_base(self):
        """update the base, including the URL for GitLab and the API endpoint."""
        self.base = self._get_and_update_setting(
            "SREGISTRY_GITLAB_BASE", "https://gitlab.com/"
        )
        self.api_base = "%s/api/v4" % self.base.strip("/")
        self.artifacts = self._get_and_update_setting(
            "SREGISTRY_GITLAB_FOLDER", "build"
        )

        self.job = self._get_and_update_setting("SREGISTRY_GITLAB_JOB", "build")

        bot.debug("      Api: %s" % self.api_base)
        bot.debug("Artifacts: %s" % self.artifacts)
        bot.debug("      Job: %s" % self.job)

    def _update_secrets(self):
        """update secrets will update metadata needed for pull and search"""
        self.token = self._required_get_and_update("SREGISTRY_GITLAB_TOKEN")
        if self.headers is None:
            self.headers = {}
        self.headers["Private-Token"] = self.token

    def __str__(self):
        return type(self)

    def _get_metadata(self):
        """since the user needs a job id and other parameters, save this
        for them.
        """
        metadata = {
            "SREGISTRY_GITLAB_FOLDER": self.artifacts,
            "api_base": self.api_base,
            "SREGISTRY_GITLAB_BASE": self.base,
            "SREGISTRY_GITLAB_JOB": self.job,
        }
        return metadata

    def artifact_to_tag(self, filename):
        """since the artifacts represent tags in GitLab (e.g., Singularity
        converts to latest, others in format Singularity.<tag>.simg
        convert to <tag> this function provides a consistent means to do
        that conversion across the library.

        Parameters
        ==========
        filename: the filename in storage to return
        """
        if filename == "Singularity.simg":
            return "latest"
        return re.sub(".*Singularity[.]", "", filename).replace(".simg", "")

    def _parse_image_name(self, image, retry=True):
        """starting with an image string in either of the following formats:
        job_id|collection
        job_id|collection|job_name

        Parse the job_name, job_id, and collection uri from it. If the user
        provides the first option, we use the job_name set by the client
        (default is build).

        Parameters
        ==========
        image: the string to parse, with values separated by |
        retry: the client can call itself recursively once, providing the
               default job_name if the user doesn't.
        """
        try:
            job_id, collection, job_name = image.split(",")
        except:
            # Retry and add job_name
            if retry:
                return self._parse_image_name("%s,%s" % (image, self.job), retry=False)

            # Or fail
            bot.exit(
                """Malformed image string! Please provide:
                        job_id,collection           (or)
                        job_id,collection,job_name"""
            )

        return job_id, collection, job_name


# Add your different functions imported at the top to the client here
Client.pull = pull

# Query functions
Client.search = search
Client._search_all = search_all



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
from sregistry.utils import parse_image_name, remove_uri
import os


def pull(self, images, file_name=None, save=True, force=False, **kwargs):
    """pull an image from an endpoint

    Parameters
    ==========
    images: refers to the uri given by the user to pull in the format
    <collection>/<namespace>. You should have an API that is able to
    retrieve a container based on parsing this uri.
    file_name: the user's requested name for the file. It can
           optionally be None if the user wants a default.
    save: if True, you should save the container to the database
      using self.add()

    Returns
    =======
    finished: a single container path, or list of paths
    """

    if not isinstance(images, list):
        images = [images]

    bot.debug("Execution of PULL for %s images" % len(images))

    finished = []
    for image in images:
        q = parse_image_name(remove_uri(image), lowercase=False)

        # Verify image existence, and obtain id
        url = "%s/container/%s/%s:%s" % (
            self.base,
            q["collection"],
            q["image"],
            q["tag"],
        )

        # Add the version, if provided
        if q["version"] is not None:
            url = "%s@%s" % (url, q["version"])

        bot.debug("Retrieving manifest at %s" % url)

        manifest = self._get(url)
        manifest["selfLink"] = url

        # If the manifest reveals a version, update names
        if "version" in manifest:
            q = parse_image_name(remove_uri(image), version=manifest["version"])

        if file_name is None:
            file_name = self._get_storage_name(q)
        file_name = os.path.abspath(file_name)

        # Determine if the user already has the image
        if os.path.exists(file_name) and not force:
            bot.exit("Image exists! Remove first, or use --force to overwrite")

        show_bar = not bool(self.quiet)
        image_file = self.download(
            url=manifest["image"],
            file_name=os.path.basename(file_name),
            show_progress=show_bar,
        )

        # If the user is saving to local storage
        if save is True:
            image_uri = "%s:%s@%s" % (
                manifest["name"],
                manifest["tag"],
                manifest["version"],
            )
            container = self.add(
                image_path=image_file,
                image_uri=image_uri,
                image_name=file_name,
                metadata=manifest,
                url=manifest["image"],
            )
            image_file = container.image

        if os.path.exists(image_file):
            bot.debug("Retrieved image file %s" % image_file)
            bot.custom(prefix="Success!", message=image_file)
            finished.append(image_file)

        # Reset file name back to None in case of multiple downloads
        file_name = None

    # If the user is only asking for one image
    if len(finished) == 1:
        finished = finished[0]
    return finished



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from spython.main import Client as Singularity
from sregistry.logger import bot
from sregistry.utils import get_tmpdir, parse_image_name, remove_uri, extract_tar
import shutil
import os


def pull(self, images, file_name=None, save=True, force=False, base=None, **kwargs):
    """pull an image from a docker hub. This is a (less than ideal) workaround
       that actually does the following:

       - creates a sandbox folder
       - adds docker layers, metadata folder, and custom metadata to it
       - converts to a squashfs image with build

    the docker manifests are stored with registry metadata.

    Parameters
    ==========
    images: refers to the uri given by the user to pull in the format
    <collection>/<namespace>. You should have an API that is able to
    retrieve a container based on parsing this uri.
    file_name: the user's requested name for the file. It can
               optionally be None if the user wants a default.
    save: if True, you should save the container to the database
          using self.add()
    base: the registry base, in case the client doesn't want to set in env.

    Returns
    =======
    finished: a single container path, or list of paths

    """

    if not isinstance(images, list):
        images = [images]

    bot.debug("Execution of PULL for %s images" % len(images))

    # If used internally we want to return a list to the user.
    finished = []
    for image in images:
        # 0. Update the base in case we aren't working with default
        base = self._update_base(image)
        q = parse_image_name(remove_uri(image), base=base)

        image_file = self._pull(
            file_name=file_name, save=save, force=force, names=q, kwargs=kwargs
        )

        finished.append(image_file)

    if len(finished) == 1:
        finished = finished[0]
    return finished


def _pull(self, file_name, names, save=True, force=False, uri="docker://", **kwargs):
    """pull an image from a docker hub. This is a (less than ideal) workaround
       that actually does the following:

       - creates a sandbox folder
       - adds docker layers, metadata folder, and custom metadata to it
       - converts to a squashfs image with build

    the docker manifests are stored with registry metadata.

    Parameters
    ==========
    images: refers to the uri given by the user to pull in the format
    <collection>/<namespace>. You should have an API that is able to
    retrieve a container based on parsing this uri.
    file_name: the user's requested name for the file. It can
               optionally be None if the user wants a default.
    save: if True, you should save the container to the database
          using self.add()

    Returns
    =======
    finished: a single container path, or list of paths
    """

    # Use Singularity to build the image, based on user preference
    if file_name is None:
        file_name = self._get_storage_name(names)

    # Determine if the user already has the image
    if os.path.exists(file_name) and force is False:
        bot.exit("Image exists! Remove first, or use --force to overwrite")

    digest = names["version"] or names["tag"]

    # Build from sandbox, prefix with sandbox
    sandbox = get_tmpdir(prefix="sregistry-sandbox")

    # First effort, get image via Sregistry
    layers = self._download_layers(names["url"], digest)

    # This is the url where the manifests were obtained
    url = self._get_manifest_selfLink(names["url"], digest)

    # Add environment to the layers
    envtar = self._get_environment_tar()
    layers = [envtar] + layers

    # Create singularity image from an empty folder
    for layer in layers:
        bot.info("Exploding %s" % layer)
        result = extract_tar(layer, sandbox, handle_whiteout=True)
        if result["return_code"] != 0:
            bot.exit(result["message"])

    sudo = kwargs.get("sudo", False)

    # Build from a sandbox (recipe) into the image_file (squashfs)
    image_file = Singularity.build(image=file_name, recipe=sandbox, sudo=sudo)

    # Fall back to using Singularity
    if image_file is None:
        bot.info("Downloading with native Singularity, please wait...")
        image = file_name.replace("docker://", uri)
        image_file = Singularity.pull(image, pull_folder=sandbox)

    # Save to local storage
    if save is True:
        # Did we get the manifests?
        manifests = {}
        if hasattr(self, "manifests"):
            manifests = self.manifests

        container = self.add(
            image_path=image_file, image_uri=names["uri"], metadata=manifests, url=url
        )

        # When the container is created, this is the path to the image
        image_file = container.image

    if os.path.exists(image_file):
        bot.debug("Retrieved image file %s" % image_file)
        bot.custom(prefix="Success!", message=image_file)

    # Clean up sandbox
    shutil.rmtree(sandbox)

    return image_file



"""

Copyright (C) 2018-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
import json
import os

################################################################################


def update_token(self):
    """update_token uses HTTP basic authentication to get a token for
    Docker registry API V2 operations. We get here if a 401 is
    returned for a request.

    Parameters
    ==========
    response: the http request response to parse for the challenge.

    https://docs.docker.com/registry/spec/auth/token/
    """
    # Add Amazon headers

    tokens = self.aws.get_authorization_token()
    token = tokens["authorizationData"][0]["authorizationToken"]

    try:
        token = {"Authorization": "Basic %s" % token}
        self.headers.update(token)

    except:
        bot.exit("Error getting token.")


def download_layers(self, repo_name, digest=None, destination=None):
    """download layers is a wrapper to do the following for a client loaded
    with a manifest for an image:

    1. use the manifests to retrieve list of digests (get_digests)
    2. atomically download the list to destination (get_layers)

    This function uses the MultiProcess client to download layers
    at the same time.
    """
    from sregistry.main.workers import Workers
    from sregistry.main.workers.aws import download_task

    # Obtain list of digets, and destination for download
    self._get_manifest(repo_name, digest)
    digests = self._get_digests(repo_name, digest)
    destination = self._get_download_cache(destination)

    # Create multiprocess download client
    workers = Workers()

    # Download each layer atomically
    tasks = []
    layers = []

    # Start with a fresh token
    self._update_token()

    for digest in digests:
        targz = "%s/%s.tar.gz" % (destination, digest["digest"])
        url = "%s/%s/blobs/%s" % (self.base, repo_name, digest["digest"])

        # Only download if not in cache already
        if not os.path.exists(targz):
            tasks.append((url, self.headers, targz))
        layers.append(targz)

    # Download layers with multiprocess workers
    if len(tasks) > 0:
        workers.run(func=download_task, tasks=tasks)

    return layers, url


def get_manifest(self, repo_name, tag):
    """return the image manifest via the aws client, saved in self.manifest"""

    image = None
    repo = self.aws.describe_images(repositoryName=repo_name)
    if "imageDetails" in repo:
        for contender in repo.get("imageDetails"):
            if tag in contender["imageTags"]:
                image = contender
                break

    # if the image isn't found, we need to exit
    if image is None:
        bot.exit("Cannot find %s:%s, is the uri correct?" % (repo_name, tag))

    digest = image["imageDigest"]
    digests = self.aws.batch_get_image(
        repositoryName=repo_name, imageIds=[{"imageDigest": digest, "imageTag": tag}]
    )

    self.manifest = json.loads(digests["images"][0]["imageManifest"])
    return self.manifest


def get_digests(self, repo_name, tag):
    """
    return a list of layers from a manifest.
    The function is intended to work with both version
    1 and 2 of the schema. All layers (including redundant)
    are returned. By default, we try version 2 first,
    then fall back to version 1.

    For version 1 manifests: extraction is reversed

    Parameters
    ==========
    manifest: the manifest to read_layers from

    """
    if not hasattr(self, "manifest"):
        bot.exit("Please retrieve manifest for the image first.")

    # version 2 manifest here!
    return self.manifest["layers"]



"""

Copyright (C) 2019-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.auth import get_credential_cache
from sregistry.defaults import SREGISTRY_DATABASE


def get_client(quiet=False, init=True, **kwargs):
    """
    get the Google Build client.

    Parameters
    ==========
    quiet: if True, suppress most output about the client (e.g. speak)

    """
    from . import Client

    Client.client_name = "google-build"
    Client.quiet = quiet

    # The user can provide environment variables via extra arguments here
    Client.envars = kwargs

    # Create credentials cache, if it doesn't exist
    Client._credential_cache = get_credential_cache()

    # Add the database, if wanted
    if SREGISTRY_DATABASE is not None:
        # These are global functions used across modules
        from sregistry.database import (
            init_db,
            add,
            cp,
            get,
            mv,
            rm,
            images,
            inspect,
            rename,
            get_container,
            get_collection,
            get_or_create_collection,
        )

        # Actions
        Client._init_db = init_db
        Client.add = add
        Client.cp = cp
        Client.get = get
        Client.inspect = inspect
        Client.mv = mv
        Client.rename = rename
        Client.rm = rm
        Client.images = images

        # Collections
        Client.get_or_create_collection = get_or_create_collection
        Client.get_container = get_container
        Client.get_collection = get_collection

    # If no database, import dummy functions that return the equivalent
    else:
        from sregistry.database import add, init_db

        Client.add = add
        Client._init_db = init_db

    # Initialize the database
    cli = Client(init=init)

    if hasattr(Client, "_init_db"):
        cli._init_db(SREGISTRY_DATABASE)
    return cli



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from retrying import retry

from sregistry.logger import bot
from sregistry.utils import confirm_delete


def delete(self, image, force=False):
    """delete an image from Google Storage.

    Parameters
    ==========
    name: the name of the file (or image) to delete

    """

    bot.debug("DELETE %s" % image)
    files = self._container_query(image)

    for file_object in files:
        if confirm_delete(file_object.name, force):
            file_object.delete()


@retry(
    wait_exponential_multiplier=1000,
    wait_exponential_max=10000,
    stop_max_attempt_number=3,
)
def destroy(self, name):
    """destroy an instance, meaning take down the instance and stop the build.
    Parameters
    ==========
    name: the name of the instance to stop building.
    """

    instances = self._get_instances()
    project = self._get_project()
    zone = self._get_zone()
    found = False

    if "items" in instances:
        for instance in instances["items"]:
            if instance["name"] == name:
                found = True
                break

    if found:
        bot.info("Killing instance %s" % name)
        return (
            self._compute_service.instances()
            .delete(project=project, zone=zone, instance=name)
            .execute()
        )



"""

Copyright (C) 2018-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
from sregistry.utils import parse_image_name, remove_uri
import os


def pull(self, images, file_name=None, save=True, **kwargs):
    """pull an image from gitlab. The image is found based on the
    uri that should correspond to a gitlab repository, and then
    the branch, job name, artifact folder, and tag of the container.
    The minimum that we need are the job id, collection, and job name. Eg:

    job_id|collection|job_name   (or)
    job_id|collection

    Parameters
    ==========
    images: refers to the uri given by the user to pull in the format
            specified above
    file_name: the user's requested name for the file. It can
               optionally be None if the user wants a default.
    save: if True, you should save the container to the database
          using self.add()

    Returns
    =======
    finished: a single container path, or list of paths

    """
    force = False
    if "force" in kwargs:
        force = kwargs["force"]

    if not isinstance(images, list):
        images = [images]

    bot.debug("Execution of PULL for %s images" % len(images))

    # If used internally we want to return a list to the user.
    finished = []
    for image in images:
        # Format job_id|collection|job_name
        # 122056733,singularityhub/gitlab-ci'
        # 122056733,singularityhub/gitlab-ci,build

        job_id, collection, job_name = self._parse_image_name(image)
        names = parse_image_name(remove_uri(collection))

        # If the user didn't provide a file, make one based on the names
        if file_name is None:
            file_name = self._get_storage_name(names)

        # If the file already exists and force is False
        if os.path.exists(file_name) and force is False:
            bot.exit("Image exists! Remove first, or use --force to overwrite")

        # Put together the GitLab URI
        image_name = "Singularity.%s.simg" % (names["tag"])
        if names["tag"] == "latest":
            image_name = "Singularity.simg"

        # Assemble artifact path
        artifact_path = "%s/%s" % (self.artifacts, image_name)
        bot.info(
            "Looking for artifact %s for job name %s, %s"
            % (artifact_path, job_name, job_id)
        )

        # project = quote_plus(collection.strip('/'))

        # This is supposed to work, but it doesn't
        # url = "%s/projects/%s/jobs/%s/artifacts/file/%s" %(self.api_base,
        #                                                  project, job_id,
        #                                                  artifact_path)

        # This does work :)
        url = "%s/%s/-/jobs/%s/artifacts/raw/%s/?inline=false" % (
            self.base,
            collection,
            job_id,
            artifact_path,
        )

        bot.info(url)

        # stream the url content to the file name
        image_file = self.download(url=url, file_name=file_name, show_progress=True)

        metadata = self._get_metadata()
        metadata["collection"] = collection
        metadata["job_id"] = job_id
        metadata["job_name"] = job_name
        metadata["artifact_path"] = artifact_path
        metadata["sregistry_pull"] = image

        # If we save to storage, the uri is the dropbox_path
        if save is True:
            container = self.add(
                image_path=image_file, image_uri=image, metadata=metadata, url=url
            )

            # When the container is created, this is the path to the image
            image_file = container.image

        if os.path.exists(image_file):
            bot.debug("Retrieved image file %s" % image_file)
            bot.custom(prefix="Success!", message=image_file)
            finished.append(image_file)

    if len(finished) == 1:
        finished = finished[0]
    return finished



"""

Copyright (C) 2017-2021 Vanessa Sochat.

This Source Code Form is subject to the terms of the
Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed
with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""

from sregistry.logger import bot
from sregistry.utils import parse_image_name, remove_uri, confirm_delete


def delete(self, image, force=False):
    """delete an image to Singularity Registry"""

    q = parse_image_name(remove_uri(image))

    # If the registry is provided in the uri, use it
    if q["registry"] is None:
        q["registry"] = self.base

    # If the base doesn't start with http or https, add it
    q = self._add_https(q)

    url = "%s/container/%s/%s:%s" % (
        q["registry"],
        q["collection"],
        q["image"],
        q["tag"],
    )

    SREGISTRY_EVENT = self.authorize(request_type="delete", names=q)
    headers = {"Authorization": SREGISTRY_EVENT}
    self._update_headers(fields=headers)

    if confirm_delete(q["uri"], force) is True:
        response = self._delete(url)
        message = self._read_response(response)
        bot.info("Response %s, %s" % (response.status_code, message))
        # add some error handling here??
    else:
        bot.info("Delete cancelled.")

    return image
